from typing import Any, Dict, List, Optional, Sequence, TypeVar

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType
from ._data_structure_io import DataStructureIO

T = TypeVar("T", bound=Dict)  # type: ignore[type-arg]  # TODO: Higher-Kinded TypeVars


class DictIO(DataStructureIO):
    """Implementation for dicts."""

    @staticmethod
    def handles_type(arg: Any) -> bool:
        return isinstance(arg, dict)

    @staticmethod
    def extract(translatable: T, names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        return {name: translatable[name] for name in names}

    @staticmethod
    def insert(
        translatable: T, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> Optional[T]:
        translatable = dict(translatable) if copy else translatable  # type: ignore

        for name in filter(translatable.__contains__, names):
            translatable[name] = type(translatable[name])(map(tmap[name].get, translatable[name]))

        return translatable if copy else None



import sys
from typing import Any, Dict

from rics._internal_support.types import PathLikeType
from rics.misc import interpolate_environment_variables

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib  # pragma: no cover


def load_toml_file(
    path: PathLikeType,
    *,
    allow_interpolation: bool = False,
    allow_nested: bool = False,
    allow_blank: bool = False,
) -> Dict[str, Any]:
    """Load a TOML file.

    This function reads a TOML file with forced `UTF-8` encoding (as per the standard). It will optionally perform
    environment variable value interpolation as well and replace matching names in the file (in-memory, the file will
    not be changed or read more than once).

    For details about the interpolation, see :func:`rics.misc.interpolate_environment_variables`.

    Args:
        path: Path to file.
        allow_interpolation: If ``True``, perform env var value interpolation.
        allow_blank: If ``False``, blank values are considered missing.
        allow_nested: If ``False``, raise an error if variables are defined within the default value of other variables.

    Returns:
        A dict parsed from `path`.
    """
    with open(path, encoding="UTF-8") as f:
        toml_string = f.read()

    if allow_interpolation:
        toml_string = interpolate_environment_variables(toml_string, allow_nested=allow_nested, allow_blank=allow_blank)

    return tomllib.loads(toml_string)



"""Fetching of translation data.

Composite:
    * :class:`.MultiFetcher`: Solution for using multiple simple fetchers, e.g. multiple databases or file-system
      locations. Or a combination thereof!

Simple fetchers:
    * :class:`.SqlFetcher`: Fetching from a single SQL database or schema.
    * :class:`.PandasFetcher`: File-system fetching based on pandas read-functions. Valid URL schemes include http, ftp,
      s3, gs, and file.
    * :class:`.MemoryFetcher`: In-memory solution, used primarily for testing.

Base fetchers:
    * :class:`.Fetcher`: Top-level interface definition. Base for all fetching implementations.
    * :class:`.AbstractFetcher`: Implements high-level operations such as
      `placeholder mapping <../documentation/translation-primer.html#placeholder-mapping>`__.
"""
from ._abstract_fetcher import AbstractFetcher
from ._cache import CacheAccess, CacheMetadata
from ._fetcher import Fetcher
from ._memory_fetcher import MemoryFetcher
from ._multi_fetcher import MultiFetcher
from ._pandas_fetcher import PandasFetcher
from ._sql_fetcher import SqlFetcher

__all__ = [
    "Fetcher",
    "AbstractFetcher",
    "CacheAccess",
    "CacheMetadata",
    "MemoryFetcher",
    "MultiFetcher",
    "PandasFetcher",
    "SqlFetcher",
]



from typing import Any, Callable, Dict, List, Optional, Sequence, TypeVar

import numpy as np
import pandas as pd
from rics.collections.misc import as_list

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType
from . import DataStructureIO
from .exceptions import NotInplaceTranslatableError

T = TypeVar("T", list, np.ndarray, tuple, pd.Index)  # type: ignore[type-arg]  # TODO: Higher-Kinded TypeVars


class SequenceIO(DataStructureIO):
    """Implementation for numpy arrays, Python lists and tuples."""

    @staticmethod
    def handles_type(arg: Any) -> bool:
        return isinstance(arg, (list, np.ndarray, tuple, pd.Index))

    @staticmethod
    def extract(translatable: T, names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        verify_names(len(translatable), names)
        return (
            {names[0]: as_list(translatable)}
            if len(names) == 1
            else {n: as_list(r) for n, r in zip(names, translatable)}
        )

    @staticmethod
    def insert(
        translatable: T, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> Optional[T]:
        verify_names(len(translatable), names)
        t = translate_sequence(translatable, names, tmap)

        ctor: Callable[[List[Optional[str]]], T]
        if copy:
            if isinstance(translatable, pd.Index):
                ctor = pd.Index
            elif isinstance(translatable, np.ndarray):
                ctor = np.array
            else:
                ctor = type(translatable)
            return ctor(t)

        try:
            translatable[:] = t[:]  # type: ignore
            return None
        except TypeError as e:
            raise NotInplaceTranslatableError(translatable) from e


def translate_sequence(
    s: T, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType]
) -> List[Optional[str]]:
    """Return a translated copy of the sequence `s`."""
    if len(names) == 1:
        return list(map(tmap[names[0]].get, s))

    return [
        translate_sequence(element, [name], tmap)  # type: ignore
        if SequenceIO.handles_type(element)
        else tmap[name].get(element)
        for name, element in zip(names, s)
    ]


def verify_names(data_len: int, names: List[NameType]) -> None:  # pragma: no cover
    """Verify that the length of names is either 1 or equal to the length of the data."""
    num_names = len(names)
    if num_names != 1 and num_names != data_len:
        raise ValueError(
            f"Number of names {len(names)} must be 1 or equal to the length of the data ({data_len}) to "
            f"translate, but got {names=}."
        )



"""Types used for translation.

This module cannot be called just `types` as that will make MyPY complain.
"""
import abc as _abc
import typing as _type
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas  # noqa: F401
    from numpy.typing import NDArray

Translatable = _type.TypeVar(
    "Translatable",
    # Primitive types
    str,
    int,
    _type.Dict,  # type: ignore[type-arg]  # TODO: Need Higher-Kinded TypeVars
    _type.Sequence,  # type: ignore[type-arg]  # TODO: Need Higher-Kinded TypeVars
    "NDArray",  # type: ignore[type-arg]  # TODO: Need Higher-Kinded TypeVars
    "pandas.DataFrame",
    "pandas.Index",
    "pandas.Series",
)
"""Enumeration of translatable types.

The only truly :attr:`Translatable` types are ``int`` and ``str``. Working with single IDs is rarely desirable in
practice, so collections such as sequences (lists, tuples, :mod:`numpy` arrays) and dict values of sequences of
primitives (or plain primitives) may be translated as well. Special handling is also implemented for :mod:`pandas`
types.
"""

ID: str = "id"
"""Name of the ID placeholder."""

NameType = _type.TypeVar("NameType", bound=_type.Hashable)
"""Type used to label collections of IDs, such as the column names in a DataFrame or the keys of a dict."""

IdType = _type.TypeVar("IdType", int, str)
"""Type of the value being translated into human-readable labels."""

SourceType = _type.TypeVar("SourceType", bound=_type.Hashable)
"""Type used to describe sources. Typically a string for things like files and database tables."""

NamesPredicate = _type.Callable[[NameType], bool]
"""A predicate type on names."""
NameTypes = _type.Union[NameType, _type.Iterable[NameType]]
"""A union of a name type, or an iterable thereof."""
Names = _type.Union[NameTypes[NameType], NamesPredicate[NameType]]
"""Acceptable name types."""


class HasSources(_abc.ABC, _type.Generic[SourceType]):
    """Indicates that `sources` and `placeholders` are available."""

    @property
    def sources(self) -> _type.List[SourceType]:
        """A list of known Source names, such as ``cities`` or ``languages``."""
        return list(self.placeholders)

    @property
    @_abc.abstractmethod
    def placeholders(self) -> _type.Dict[SourceType, _type.List[str]]:
        """Placeholders for all known Source names, such as ``id`` or ``name``.

        These are the (possibly unmapped) placeholders that may be used for translation.

        Returns:
            A dict ``{source: [placeholders..]}``.
        """  # noqa: DAR202



"""Offline (in-memory) translation classes."""
from ._format import Format
from ._format_applier import DefaultFormatApplier, FormatApplier
from ._magic_dict import MagicDict
from ._translation_map import TranslationMap

__all__ = [
    "Format",
    "FormatApplier",
    "DefaultFormatApplier",
    "TranslationMap",
    "MagicDict",
]



import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Mapping, Union

import pandas as pd
from rics._internal_support.types import PathLikeType
from rics.misc import get_by_full_name, tname

from ..offline.types import PlaceholderTranslations
from ..types import IdType
from ._abstract_fetcher import AbstractFetcher
from .types import FetchInstruction

PandasReadFunction = Callable[[PathLikeType, Any, Any], pd.DataFrame]
FormatFn = Callable[[PathLikeType], str]


class PandasFetcher(AbstractFetcher[str, IdType]):
    """Fetcher implementation using pandas ``DataFrame`` s as the data format.

    Fetch data from serialized ``DataFrame`` s. How this is done is determined by the `read_function`. This is typically
    a Pandas function such as :func:`pandas.read_csv` or :func:`pandas.read_pickle`, but any function that accepts a
    string `source` as the  first argument and returns a data frame can be used.

    Args:
        read_function: A Pandas `read`-function. If a string is given, the function is resolved using
            func:`rics.misc.get_by_full_name`. Unqualified named are assumed to belong to the ``pandas`` namespace.
        read_path_format: A formatting string or a callable to apply to a source before passing them to `read_function`.
            Must contain a `source` as its only placeholder. Example: ``data/{source}.pkl``. Leave as-is if ``None``.
            Valid URL schemes include http, ftp, s3, gs, and file.
        read_function_kwargs: Additional keyword arguments for `read_function`.
        online: Setting ``online=False`` typically indicates that files are hosted at a location where there are access
            limitations, e.g. through data transfer fees.

    See Also:
        The official `Pandas IO documentation <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html>`_
    """

    def __init__(
        self,
        read_function: Union[PandasReadFunction, str] = "read_pickle",
        read_path_format: Union[str, FormatFn] = "data/{}.pkl",
        read_function_kwargs: Mapping[str, Any] = None,
        online: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._read = get_by_full_name(read_function, pd) if isinstance(read_function, str) else read_function
        self._format_source: FormatFn  # Oneliner makes MyPy complain?
        if callable(read_path_format):
            self._format_source = read_path_format  # pragma: no cover
        else:
            self._format_source = read_path_format.format
        self._online = online
        self._kwargs = read_function_kwargs or {}

        self._source_paths: Dict[str, Path] = {}
        self._sources: List[str] = []
        self._placeholders: Dict[str, List[str]] = {}

    def read(self, source_path: PathLikeType) -> pd.DataFrame:
        """Read a ``DataFrame`` from a source path.

        Args:
            source_path: Path to serialized ``DataFrame``.

        Returns:
            A deserialized ``DataFrame``.
        """
        return self._read(source_path, **self._kwargs)

    def find_sources(self) -> Dict[str, Path]:
        """Search for source paths to pass to `read_function` using `read_path_format`.

        Returns:
            A dict ``{source, path}``.

        Raises:
            IOError: If files cannot be read.
        """
        abs_file = Path(self._format_source("")).absolute()
        directory = abs_file.parent
        file_pattern = abs_file.name

        if not directory.is_dir():  # pragma: no cover
            problem = "is not a directory" if directory.exists() else "does not exist"
            raise IOError(f"Bad path format: {directory} {problem}.")

        source_paths = {}
        # Path.glob does not work with absolute directories.
        for file in map(Path, os.listdir(directory)):
            if file.name.endswith(str(file_pattern)):  # pragma: no cover
                source_paths[file.name.replace(file_pattern, "")] = directory.joinpath(file)

        if not source_paths:  # pragma: no cover
            pattern = Path(self._format_source("*")).absolute()
            raise IOError(f"Bad path pattern: '{pattern}' did not match any files.")

        return source_paths

    @property
    def sources(self) -> List[str]:
        if not self._sources:  # pragma: no cover
            if not self._source_paths:
                self._source_paths = self.find_sources()
            self._sources = list(self._source_paths)
            self.logger.getChild("pandas").getChild("discovery").debug("Sources initialized: %s", self._sources)

        return self._sources

    @property
    def placeholders(self) -> Dict[str, List[str]]:
        if not self._placeholders:
            self._placeholders = {
                source: list(self.read(self._source_paths[source]).columns) for source in self.sources
            }

        return self._placeholders

    def fetch_translations(self, instr: FetchInstruction[str, IdType]) -> PlaceholderTranslations[str]:
        return PlaceholderTranslations.make(
            instr.source,
            self.read(self._source_paths[instr.source]),
        )

    @property
    def online(self) -> bool:
        return self._online

    def __repr__(self) -> str:
        read_path_format = self._format_source("{source}")
        return f"{tname(self)}(read_function={tname(self._read)}, {read_path_format=})"



from abc import abstractmethod
from typing import Generic, Iterable

from ..offline.types import SourcePlaceholderTranslations
from ..types import HasSources, IdType, SourceType
from .types import IdsToFetch


class Fetcher(Generic[SourceType, IdType], HasSources[SourceType]):
    """Interface for fetching translations from an external source."""

    @property
    @abstractmethod
    def allow_fetch_all(self) -> bool:
        """Flag indicating whether the :meth:`~.fetching.Fetcher.fetch_all` operation is permitted."""

    def close(self) -> None:
        """Close the ``Fetcher``. Does nothing by default."""

    @property
    @abstractmethod
    def online(self) -> bool:
        """Return connectivity status. If ``False``, no new translations may be fetched."""

    @abstractmethod
    def fetch(
        self,
        ids_to_fetch: Iterable[IdsToFetch[SourceType, IdType]],
        placeholders: Iterable[str] = (),
        required: Iterable[str] = (),
    ) -> SourcePlaceholderTranslations[SourceType]:
        """Retrieve placeholder translations from the source.

        Args:
            ids_to_fetch: Tuples (source, ids) to fetch. If ``ids=None``, retrieve data for as many IDs as possible.
            placeholders: All desired placeholders in preferred order.
            required: Placeholders that must be included in the response.

        Returns:
            A mapping ``{source: PlaceholderTranslations}`` for translation.

        Raises:
            UnknownPlaceholderError: For placeholder(s) that are unknown to the ``Fetcher``.
            UnknownSourceError: For sources(s) that are unknown to the ``Fetcher``.
            ForbiddenOperationError: If trying to fetch all IDs when not possible or permitted.
            ImplementationError: For errors made by the inheriting implementation.

        Notes:
            Placeholders are usually columns in relational database applications. These are the components which are
            combined to create ID translations. See :class:`~id_translation.offline.Format` documentation for details.
        """

    @abstractmethod
    def fetch_all(
        self,
        placeholders: Iterable[str] = (),
        required: Iterable[str] = (),
    ) -> SourcePlaceholderTranslations[SourceType]:
        """Fetch as much data as possible.

        Args:
            placeholders: All desired placeholders in preferred order.
            required: Placeholders that must be included in the response.

        Returns:
            A mapping ``{source: PlaceholderTranslations}`` for translation.

        Raises:
            ForbiddenOperationError: If fetching all IDs is not possible or permitted.
            UnknownPlaceholderError: For placeholder(s) that are unknown to the ``Fetcher``.
            ImplementationError: For errors made by the inheriting implementation.
        """



"""Translation of IDs with flexible formatting and name matching.

For an introduction to translation, see :ref:`translation-primer` and :ref:`mapping-primer`.
"""

import logging

from ._base_metadata import BaseMetadata
from ._config_utils import ConfigMetadata
from ._load_toml import load_toml_file
from ._translator import Translator
from .factory import TranslatorFactory

from .__version__ import __author__  # isort:skip
from .__version__ import __copyright__  # isort:skip
from .__version__ import __title__, __description__, __version__  # isort:skip

__all__ = [
    "Translator",
    "TranslatorFactory",
    "BaseMetadata",
    "ConfigMetadata",
    "load_toml_file",
    "__version__",  # Make MyPy happy
]

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Integration for insertion and extraction of IDs and translations to and from various data structures."""

from ._data_structure_io import DataStructureIO
from ._resolve import resolve_io

__all__ = ["DataStructureIO", "resolve_io"]



from typing import Any, Dict, Generic, Hashable, Iterable, Mapping, Optional, Tuple, TypeVar

from rics.misc import tname

from ._cardinality import Cardinality
from .exceptions import CardinalityError
from .types import HL, HR, LeftToRight, RightToLeft

HAnySide = TypeVar("HAnySide", bound=Hashable)
MatchTupleAnySide = TypeVar("MatchTupleAnySide", bound=Hashable)  # TODO: Higher-Kinded TypeVars


class DirectionalMapping(Generic[HL, HR]):
    """A two-way mapping between hashable elements.

    Args:
        cardinality: Explicit cardinality. Derive if ``None``.
        left_to_right: A left-to-right mapping of elements.
        right_to_left: A right-to-left mapping of elements.
        _verify: If ``False``, input checks are disabled. Intended for internal use.

    Raises:
        ValueError: If both of `left_to_right` and `right_to_left` are ``None``.
        ValueError: If verification of two-sided input fails, and ``verify=True``.
        CardinalityError: If explicit `cardinality` < :attr:`cardinality`, and ``verify=True``.
    """

    def __init__(
        self,
        cardinality: Cardinality.ParseType = None,
        left_to_right: Mapping[HL, Iterable[HR]] = None,
        right_to_left: Mapping[HR, Iterable[HL]] = None,
        _verify: bool = True,
    ) -> None:
        self._left_to_right: Dict[HL, Tuple[HR, ...]] = self._to_other(left_to_right, right_to_left)
        self._right_to_left: Dict[HR, Tuple[HL, ...]] = self._to_other(right_to_left, left_to_right)

        if left_to_right is not None and right_to_left is not None and _verify:
            self._verify(expected=DirectionalMapping(cardinality, left_to_right=left_to_right, _verify=False))

        self._cardinality = self._handle_cardinality(cardinality, self._left_to_right, self._right_to_left, _verify)

    @property
    def cardinality(self) -> Cardinality:
        """Cardinality with which this mapping was created.

        Returns:
            Cardinality with which this mapping was created.
        """
        return self._cardinality

    @property
    def left(self) -> Tuple[HL, ...]:
        """Left-side elements in the mapping."""
        return tuple(self._left_to_right)

    @property
    def right(self) -> Tuple[HR, ...]:
        """Right-side elements in the mapping."""
        return tuple(self._right_to_left)

    @property
    def left_to_right(self) -> LeftToRight[HL, HR]:
        """Left-to-right element mappings."""
        return self._left_to_right

    @property
    def right_to_left(self) -> RightToLeft[HR, HL]:
        """Right-to-left element mappings."""
        return self._right_to_left

    @property
    def reverse(self) -> "DirectionalMapping[HR, HL]":
        """Reverse the mapping by swapping the sides.

        Returns:
            A copy with data identical to the calling instance, but with sides inversed compared to the caller.
        """
        return DirectionalMapping(
            self.cardinality.inverse,
            left_to_right=self._right_to_left.copy(),
            right_to_left=self._left_to_right.copy(),
            _verify=False,
        )

    def flatten(self) -> Dict[HL, HR]:
        """Return a flattened version of self as a dict.

        Returns:
            A dict ``{left: right}``.

        Raises:
            CardinalityError: If cardinality is not :attr:`~.Cardinality.OneToOne` or :attr:`~.Cardinality.ManyToOne`.
        """
        if not self._cardinality.one_right:  # pragma: no cover
            raise CardinalityError(f"Must have one of {(Cardinality.OneToOne, Cardinality.ManyToOne)} to flatten.")

        return {left: right[0] for left, right in self._left_to_right.items()}

    def select_left(self, elements: Iterable[HL], exclude: bool = False) -> "DirectionalMapping[HL, HR]":
        """Perform a selection on left-side elements.

        Args:
            elements: Elements to select.
            exclude: If ``True``, return everything **except** the given elements.

        Returns:
            A new Mapping for the selection.

        Raises:
            KeyError: If any of the chosen elements do not exist and ``exclude=False``.
        """
        return DirectionalMapping(None, left_to_right=_select(elements, self._left_to_right, exclude))

    def select_right(self, elements: Iterable[HR], exclude: bool = False) -> "DirectionalMapping[HL, HR]":
        """Perform a selection on right-side elements.

        Args:
            elements: Elements to select.
            exclude: If ``True``, return everything **except** the given elements.

        Returns:
            A new instance for the selection.

        Raises:
            KeyError: If any of the chosen elements do not exist and ``exclude=False``.
        """
        return DirectionalMapping(None, right_to_left=_select(elements, self._right_to_left, exclude))

    @classmethod
    def _to_other(
        cls,
        primary_side: Mapping[Any, Iterable[Any]] = None,
        backup_side: Mapping[Any, Iterable[Any]] = None,
    ) -> Dict[Any, Any]:
        if primary_side is not None:
            return primary_side  # type: ignore[return-value]

        if backup_side is None:
            raise ValueError("At least one side must be given")

        from collections import defaultdict

        other_side = defaultdict(list)
        for k, matches_for_k in backup_side.items():
            for m in matches_for_k:
                other_side[m].append(k)
        return {k: tuple(set(m)) for k, m in other_side.items()}

    def _verify(self, expected: "DirectionalMapping[HL, HR]") -> None:
        # TODO; This should only be done during testing I think
        self._verify_side(self.left, expected.left, "Left")
        self._verify_side(self.right, expected.right, "Right")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DirectionalMapping):
            return False

        return (
            self._left_to_right == other._left_to_right
            and self._right_to_left == other._right_to_left
            and self._cardinality == other._cardinality
        )

    def __repr__(self) -> str:
        n_left = len(self._left_to_right)
        n_right = len(self._right_to_left)
        return f"{tname(self)}({n_left} left | {n_right} right, type={self.cardinality.name})"

    def __bool__(self) -> bool:
        return bool(self._left_to_right or self._right_to_left)

    @classmethod
    def _handle_cardinality(
        cls,
        expected: Optional[Cardinality.ParseType],
        left: LeftToRight[HL, HR],
        right: RightToLeft[HR, HL],
        verify: bool,
    ) -> Cardinality:
        if not (left and right):
            if expected is None:
                return Cardinality.ManyToMany
            else:
                return Cardinality.parse(expected)

        actual = Cardinality.from_counts(
            left_count=max(map(len, right.values())), right_count=max(map(len, left.values()))
        )

        if expected is None:
            return actual
        else:
            expected = Cardinality.parse(expected)

        if verify and actual > expected:
            raise CardinalityError(
                f"Cannot cast explicit given type {expected} to actual type {actual} " f"for ({left=} | {right=})",
            )

        return expected

    @classmethod
    def _verify_side(cls, actual: MatchTupleAnySide, expected: MatchTupleAnySide, name: str) -> None:
        if actual != expected:
            raise ValueError(f"{name}-side mismatch: Got {actual} but expected {expected}.")


def _select(elements: Iterable[HAnySide], items: Dict[HL, Tuple[HR, ...]], exclude: bool) -> Dict[HL, Tuple[HR, ...]]:
    if not exclude:
        missing_elements = set(elements).difference(items)
        if missing_elements:
            raise KeyError(f"Unknown keys {missing_elements} in {items}.")

    s = set(elements)
    chosen_elements = filter(lambda e: e not in s, items) if exclude else filter(s.__contains__, items)
    return {e: items[e] for e in chosen_elements}



from typing import Any, Dict, List, Optional, Sequence, TypeVar

import pandas as pd

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType
from ._data_structure_io import DataStructureIO
from ._sequence import SequenceIO, translate_sequence, verify_names

T = TypeVar("T", pd.DataFrame, pd.Series)


def _cast_series(series: pd.Series) -> pd.Series:
    return series.dropna().astype(int) if series.dtype == float else series


class PandasIO(DataStructureIO):
    """Implementation for Pandas data types."""

    @staticmethod
    def handles_type(arg: Any) -> bool:
        return isinstance(arg, (pd.DataFrame, pd.Series))

    @staticmethod
    def extract(translatable: T, names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        if isinstance(translatable, pd.DataFrame):
            return {name: _cast_series(translatable[name]).tolist() for name in names}
        else:
            return SequenceIO.extract(_cast_series(translatable), names)

    @staticmethod
    def insert(
        translatable: T, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> Optional[T]:
        translatable = translatable.copy() if copy else translatable

        if isinstance(translatable, pd.DataFrame):
            for name in names:
                translatable[name] = translatable[name].map(tmap[name].get)
        else:
            verify_names(len(translatable), names)
            translatable.update(pd.Series(translate_sequence(translatable, names, tmap), index=translatable.index))

        return translatable if copy else None



"""Insertion and extraction of IDs and translations."""
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Sequence

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType


class DataStructureIO:
    """Insertion and extraction of IDs and translations."""

    @staticmethod
    def handles_type(arg: Any) -> bool:
        """Return ``True`` if the implementation handles data for the type of `arg`."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def extract(c: Any, names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        """Extract IDs from `c`.

        Args:
            c: A collection to extract IDs from.
            names: List of names to extract IDs for.

        Returns:
            A dict ``{name, ids}``.
        """

    @staticmethod
    @abstractmethod
    def insert(
        c: Any, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> Optional[Any]:
        """Insert translations into `c`.

        Args:
            c: A collection apply translations for. Modified iff ``copy=False``.
            names: Names in `t` to translate..
            tmap: Translations for IDs in `c`.
            copy: If ``True``, modify contents of the original collection `c`. Otherwise, return a copy.

        Returns:
            A copy of `c` if ``copy=True``. ``None`` otherwise.

        Raises:
            NotInplaceTranslatableError: If ``copy=False`` for a type which is not translatable in-place.
        """



from __future__ import annotations

import logging
import warnings
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from time import perf_counter
from typing import Dict, Iterable, List, Tuple

from rics.action_level import ActionLevel, ActionLevelHelper
from rics.collections.dicts import reverse_dict
from rics.misc import tname
from rics.performance import format_seconds

from ..offline.types import SourcePlaceholderTranslations
from ..types import IdType, SourceType
from . import Fetcher, exceptions
from .types import IdsToFetch

LOGGER = logging.getLogger(__package__).getChild("MultiFetcher")

FetchResult = Tuple[int, SourcePlaceholderTranslations[SourceType]]

_ACTION_LEVEL_HELPER = ActionLevelHelper(
    duplicate_translation_action=ActionLevel.IGNORE,
    duplicate_source_discovered_action=None,
)


class MultiFetcher(Fetcher[SourceType, IdType]):
    """Fetcher which combines the results of other fetchers.

    Args:
        *fetchers: Fetchers to wrap.
        max_workers: Number of threads to use for fetching. Fetch instructions will be dispatched using a
             :py:class:`~concurrent.futures.ThreadPoolExecutor`. Individual fetchers will be called at most once per
             ``fetch()`` or ``fetch_all()`` call made with the ``MultiFetcher``.
        duplicate_translation_action: Action to take when multiple fetchers return translations for the same source.
        duplicate_source_discovered_action: Action to take when multiple fetchers claim the same source.
    """

    def __init__(
        self,
        *fetchers: Fetcher[SourceType, IdType],
        max_workers: int = 1,
        duplicate_translation_action: ActionLevel.ParseType = ActionLevel.WARN,
        duplicate_source_discovered_action: ActionLevel.ParseType = ActionLevel.WARN,
    ) -> None:
        for pos, f in enumerate(fetchers):
            if not isinstance(f, Fetcher):  # pragma: no cover
                raise TypeError(f"Argument {pos} is of type {type(f)}, expected Fetcher subtype.")

        self._id_to_rank: Dict[int, int] = {id(f): rank for rank, f in enumerate(fetchers)}
        self._id_to_fetcher: Dict[int, Fetcher[SourceType, IdType]] = {id(f): f for f in fetchers}
        self._source_to_fetcher_id_actual: Dict[SourceType, int] = {}
        self.max_workers: int = max_workers
        self._duplicate_translation_action = _ACTION_LEVEL_HELPER.verify(
            duplicate_translation_action, "duplicate_translation_action"
        )
        self._duplicate_source_discovered_action = _ACTION_LEVEL_HELPER.verify(
            duplicate_source_discovered_action, "duplicate_source_discovered_action"
        )

        if len(self.fetchers) != len(fetchers):
            raise ValueError("Repeat fetcher instance(s)!")  # pragma: no cover

    @property
    def allow_fetch_all(self) -> bool:
        return all(f.allow_fetch_all for f in self._id_to_fetcher.values())  # pragma: no cover

    @property
    def online(self) -> bool:
        return all(f.online for f in self._id_to_fetcher.values())  # pragma: no cover

    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        return {
            source: self._id_to_fetcher[self._source_to_fetcher_id[source]].placeholders[source]
            for source in self.sources
        }

    @property
    def fetchers(self) -> List[Fetcher[SourceType, IdType]]:
        """Return child fetchers."""
        return list(self._id_to_fetcher.values())

    @property
    def sources(self) -> List[SourceType]:
        return list(self._source_to_fetcher_id)

    def fetch(
        self,
        ids_to_fetch: Iterable[IdsToFetch[SourceType, IdType]],
        placeholders: Iterable[str] = (),
        required: Iterable[str] = (),
    ) -> SourcePlaceholderTranslations[SourceType]:
        tasks: Dict[int, List[IdsToFetch[SourceType, IdType]]] = {}
        sources = []
        for idt in ids_to_fetch:
            tasks.setdefault(self._source_to_fetcher_id[idt.source], []).append(idt)
            sources.append(idt.source)

        placeholders = tuple(placeholders)
        required = tuple(required)

        n_sources_and_fetchers = f"{len(sources)} sources using {len(tasks)} different fetchers"

        start = perf_counter()
        if LOGGER.isEnabledFor(logging.DEBUG):
            event_key = f"{self.__class__.__name__.upper()}.FETCH"
            LOGGER.debug(
                f"Dispatch FETCH jobs for {n_sources_and_fetchers} on {self.max_workers} threads.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    sources=sources,
                    placeholders=placeholders,
                    required_placeholders=required,
                    max_workers=self.max_workers,
                    num_fetchers=len(tasks),
                    fetch_all=False,
                ),
            )

        def fetch(fid: int) -> FetchResult[SourceType]:
            fetcher = self._id_to_fetcher[fid]
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug(f"Begin FETCH job for {len(tasks[fid])} sources using {self._fmt_fetcher(fetcher)}.")
            return fid, fetcher.fetch(tasks[fid], placeholders, required=required)

        with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix=tname(self)) as executor:
            futures = [executor.submit(fetch, fid) for fid in tasks]
            ans = self._gather(futures)

        if LOGGER.isEnabledFor(logging.DEBUG):
            execution_time = perf_counter() - start
            LOGGER.debug(
                f"Completed FETCH jobs for {n_sources_and_fetchers} in {format_seconds(execution_time)}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    execution_time=execution_time,
                    sources=len(ans),
                    max_workers=self.max_workers,
                    num_fetchers=len(tasks),
                    fetch_all=False,
                ),
            )

        return ans

    def fetch_all(
        self, placeholders: Iterable[str] = (), required: Iterable[str] = ()
    ) -> SourcePlaceholderTranslations[SourceType]:
        placeholders = tuple(placeholders)
        required = tuple(required)

        start = perf_counter()
        if LOGGER.isEnabledFor(logging.DEBUG):
            event_key = f"{self.__class__.__name__.upper()}.FETCH_ALL"
            LOGGER.debug(
                f"Dispatch FETCH_ALL jobs for {len(self.fetchers)} fetchers on {self.max_workers} threads.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    placeholders=placeholders,
                    required_placeholders=required,
                    max_workers=self.max_workers,
                    num_fetchers=len(self.fetchers),
                    fetch_all=True,
                ),
            )

        def fetch_all(fetcher: Fetcher[SourceType, IdType]) -> FetchResult[SourceType]:
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug(f"Begin FETCH_ALL job using {self._fmt_fetcher(fetcher)}.")
            return id(fetcher), fetcher.fetch_all(placeholders, required=required)

        with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix=tname(self)) as executor:
            futures = [executor.submit(fetch_all, fetcher) for fetcher in self.fetchers]
            ans = self._gather(futures)

        if LOGGER.isEnabledFor(logging.DEBUG):
            execution_time = perf_counter() - start
            LOGGER.debug(
                f"Completed FETCH_ALL jobs for {len(ans)} sources using "
                f"{len(self.fetchers)} fetchers in {format_seconds(execution_time)}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="EXIT",
                    event_title=f"{event_key}.EXIT",
                    execution_time=execution_time,
                    sources=list(ans),
                    max_workers=self.max_workers,
                    num_fetchers=len(self.fetchers),
                    fetch_all=True,
                ),
            )
        return ans

    @property
    def duplicate_translation_action(self) -> ActionLevel:
        """Return action to take when multiple fetchers return translations for the same source."""
        return self._duplicate_translation_action

    @property
    def duplicate_source_discovered_action(self) -> ActionLevel:
        """Return action to take when multiple fetchers claim the same source."""
        return self._duplicate_source_discovered_action

    @property
    def _source_to_fetcher_id(self) -> Dict[SourceType, int]:
        if not self._source_to_fetcher_id_actual:
            source_ranks: Dict[SourceType, int] = {}
            source_to_fetcher_id: Dict[SourceType, int] = {}

            for fid, fetcher in list(self._id_to_fetcher.items()):
                sources = fetcher.sources

                if not sources:
                    LOGGER.warning(f"Discarding {self._fmt_fetcher(fetcher)}: No sources.")
                    fetcher.close()
                    del self._id_to_rank[fid]
                    del self._id_to_fetcher[fid]
                    continue

                rank = self._id_to_rank[fid]
                for source in sources:
                    if source in source_to_fetcher_id:
                        self._log_rejection(source, rank, source_ranks[source], translation=False)
                    else:
                        source_to_fetcher_id[source] = fid
                        source_ranks[source] = rank
            self._source_to_fetcher_id_actual = source_to_fetcher_id

        return self._source_to_fetcher_id_actual

    def _gather(self, futures: Iterable[Future[FetchResult[SourceType]]]) -> SourcePlaceholderTranslations[SourceType]:
        ans: SourcePlaceholderTranslations[SourceType] = {}
        source_ranks: Dict[SourceType, int] = {}

        for future in as_completed(futures):
            fid, translations = future.result()
            rank = self._id_to_rank[fid]
            self._process_future_result(translations, rank, source_ranks, ans)
        return ans

    def _process_future_result(
        self,
        translations: SourcePlaceholderTranslations[SourceType],
        rank: int,
        source_ranks: Dict[SourceType, int],
        ans: SourcePlaceholderTranslations[SourceType],
    ) -> None:
        for source_translations in translations.values():
            source = source_translations.source
            other_rank = source_ranks.setdefault(source, rank)

            if other_rank != rank:
                self._log_rejection(source, rank, other_rank, translation=True)
                if rank > other_rank:
                    continue  # Don't save -- other rank is greater (lower-is-better).

            ans[source] = source_translations

    def _log_rejection(self, source: SourceType, rank0: int, rank1: int, translation: bool) -> None:  # pragma: no cover
        accepted_rank, rejected_rank = (rank0, rank1) if rank0 < rank1 else (rank1, rank0)

        rank_to_id = reverse_dict(self._id_to_rank)
        accepted = self._fmt_fetcher(self._id_to_fetcher[rank_to_id[accepted_rank]])
        rejected = self._fmt_fetcher(self._id_to_fetcher[rank_to_id[rejected_rank]])

        msg = (
            f"Discarded translations for {source=} retrieved from {rejected} since the {accepted} returned "
            "translations for the same source."
            if translation
            else f"Discarded {source=} retrieved from {rejected} since the {accepted} already claimed same source."
        )

        msg += " Hint: Rank is determined input order at initialization."

        action = self.duplicate_translation_action if translation else self.duplicate_source_discovered_action

        if action is ActionLevel.IGNORE:
            LOGGER.debug(msg)
        else:
            if action is ActionLevel.RAISE:
                LOGGER.error(msg)
                raise exceptions.DuplicateSourceError(msg)
            else:
                warnings.warn(msg, exceptions.DuplicateSourceWarning, stacklevel=2)
                LOGGER.warning(msg)

    def __repr__(self) -> str:
        max_workers = self.max_workers
        return f"{tname(self)}({max_workers=}, fetchers={', '.join(map(str, self._id_to_fetcher.values()))})"

    def _fmt_fetcher(self, fetcher: Fetcher[SourceType, IdType]) -> str:
        """Format a managed fetcher with rank and hex ID."""
        fetcher_id = id(fetcher)
        rank = self._id_to_rank[fetcher_id]
        return f"rank-{rank} fetcher {fetcher} at {hex(fetcher_id)}"



from typing import Any, Dict, List, Mapping, Sequence, Union

import pandas as pd

from ..offline.types import PlaceholderTranslations, SourcePlaceholderTranslations
from ..types import IdType, SourceType
from ._abstract_fetcher import AbstractFetcher
from .types import FetchInstruction


class MemoryFetcher(AbstractFetcher[SourceType, IdType]):
    """Fetch from memory.

    Args:
        data: A dict ``{source: PlaceholderTranslations}`` to fetch from.
    """

    def __init__(
        self,
        data: Union[
            SourcePlaceholderTranslations[SourceType],
            Dict[SourceType, PlaceholderTranslations[SourceType]],
            Dict[SourceType, pd.DataFrame],
            Mapping[SourceType, Mapping[str, Sequence[Any]]],
        ] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._data: SourcePlaceholderTranslations[SourceType] = (
            {} if not data else {source: PlaceholderTranslations.make(source, pht) for source, pht in data.items()}
        )

    @property
    def sources(self) -> List[SourceType]:
        return list(self._data)

    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        return {source: list(pht.placeholders) for source, pht in self._data.items()}

    def fetch_translations(self, instr: FetchInstruction[SourceType, IdType]) -> PlaceholderTranslations[SourceType]:
        return self._data[instr.source]



from typing import Type

from ..types import Translatable
from . import DataStructureIO
from ._dict import DictIO
from ._pandas import PandasIO
from ._sequence import SequenceIO
from ._single_value import SingleValueIO
from .exceptions import UntranslatableTypeError


def resolve_io(arg: Translatable) -> Type[DataStructureIO]:
    """Get an IO instance for `arg`.

    Args:
        arg: An argument to get IO for.

    Returns:
        A data structure IO instance for `arg`.

    Raises:
        UntranslatableTypeError: If not IO could be found.
    """
    for tio_class in DictIO, PandasIO, SequenceIO, SingleValueIO:
        if tio_class.handles_type(arg):
            return tio_class

    raise UntranslatableTypeError(type(arg))



from copy import copy
from typing import Any, Dict, Generic, Iterator, List, Mapping, Optional, Set, Type, Union

from rics.collections.dicts import InheritedKeysDict, reverse_dict
from rics.misc import tname

from ..types import HasSources, IdType, NameType, SourceType
from ._format import Format
from ._format_applier import DefaultFormatApplier, FormatApplier
from ._magic_dict import MagicDict
from .types import FormatType, SourcePlaceholderTranslations

NameToSource = Dict[NameType, SourceType]


class TranslationMap(
    Generic[NameType, SourceType, IdType],
    HasSources[SourceType],
    Mapping[Union[NameType, SourceType], MagicDict[IdType]],
):
    """Storage class for fetched translations.

    Args:
        source_translations: Fetched translations ``{source: PlaceholderTranslations}``.
        name_to_source: Mappings ``{name: source}``, but may be overridden by the user.
        fmt: A translation format. Must be given to use as a mapping.
        default_fmt: Alternative format specification to use instead of `fmt` for fallback translation.
        default_fmt_placeholders: Per-source default placeholder values.

    Notes:
        Type checking of `fmt` and `default_fmt_placeholders` attributes may fail due to
        `mypy#3004 <https://github.com/python/mypy/issues/3004>`_
    """

    FORMAT_APPLIER_TYPE: Type[FormatApplier[NameType, SourceType, IdType]] = DefaultFormatApplier
    """Format application implementation type. Overwrite attribute to customize."""

    def __init__(
        self,
        source_translations: SourcePlaceholderTranslations[SourceType],
        name_to_source: NameToSource[NameType, SourceType] = None,
        fmt: FormatType = None,
        default_fmt: FormatType = None,
        default_fmt_placeholders: InheritedKeysDict[SourceType, str, Any] = None,
    ) -> None:
        self.default_fmt = default_fmt  # type: ignore
        self.default_fmt_placeholders = default_fmt_placeholders  # type: ignore
        self.fmt = fmt  # type: ignore
        self._format_appliers: Dict[SourceType, FormatApplier[NameType, SourceType, IdType]] = {
            source: self.FORMAT_APPLIER_TYPE(translations) for source, translations in source_translations.items()
        }
        self._names_and_sources: Set[Union[NameType, SourceType]] = set()
        self.name_to_source = name_to_source or {}

        self._reverse_mode: bool = False

    def apply(
        self, name_or_source: Union[NameType, SourceType], fmt: FormatType = None, default_fmt: FormatType = None
    ) -> MagicDict[IdType]:
        """Create translations for names. Note: ``__getitem__`` delegates to this method.

        Args:
            name_or_source: A name or source to translate.
            fmt: Format to use. If ``None``, fall back to init format.
            default_fmt: Alternative format for default translation. Resolution: Arg -> init arg, fmt arg, init fmt arg

        Returns:
            Translations for `name` as a dict ``{id: translation}``.

        Raises:
            ValueError: If ``fmt=None`` and initialized without `fmt`.
            KeyError: If trying to translate `name` which is not known.
        """
        fmt = self._fmt if fmt is None else fmt
        if fmt is None:
            raise ValueError("No format specified and None given at initialization.")  # pragma: no cover

        fmt = Format.parse(fmt)
        default_fmt = self._default_fmt if default_fmt is None else Format.parse(default_fmt)
        source = self.name_to_source.get(name_or_source, name_or_source)  # type: ignore
        translations: MagicDict[IdType] = self._format_appliers[source](
            fmt,
            default_fmt=default_fmt,
            default_fmt_placeholders=self.default_fmt_placeholders.get(source),
        )
        return (
            MagicDict(
                reverse_dict(translations),  # type: ignore
                default_value=None,  # force failure for unknown keys
            )
            if self.reverse_mode
            else translations
        )

    @property
    def names(self) -> List[NameType]:
        """Return names that can be translated."""
        return list(self.name_to_source)

    @property
    def sources(self) -> List[SourceType]:
        return list(self._format_appliers)

    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        return {applier.source: applier.placeholders for applier in self._format_appliers.values()}

    @property
    def name_to_source(self) -> NameToSource[NameType, SourceType]:
        """Return name-to-source mapping."""
        return self._name_to_source

    @name_to_source.setter
    def name_to_source(self, value: NameToSource[NameType, SourceType]) -> None:
        self._name_to_source: NameToSource[NameType, SourceType] = value
        self._names_and_sources = set(value).union(self._format_appliers)

    @property
    def fmt(self) -> Optional[Format]:
        """Return the translation format."""
        return self._fmt

    @fmt.setter
    def fmt(self, value: Optional[FormatType]) -> None:
        self._fmt = None if value is None else Format.parse(value)

    @property
    def default_fmt(self) -> Optional[Format]:
        """Return the format specification to use instead of `fmt` for fallback translation."""
        return self._default_fmt

    @default_fmt.setter
    def default_fmt(self, value: Optional[FormatType]) -> None:
        self._default_fmt = None if value is None else Format.parse(value)

    @property
    def default_fmt_placeholders(self) -> InheritedKeysDict[SourceType, str, Any]:
        """Return the default translations used for `default_fmt_placeholders` placeholders."""
        return self._default_fmt_placeholders

    @default_fmt_placeholders.setter
    def default_fmt_placeholders(self, value: Optional[InheritedKeysDict[SourceType, str, Any]]) -> None:
        self._default_fmt_placeholders = InheritedKeysDict() if value is None else InheritedKeysDict.make(value)

    @property
    def reverse_mode(self) -> bool:
        """Return reversed mode status flag.

         If set, the mappings returned by :meth:`apply` (and therefore also ``__getitem__``) are reversed.

        Returns:
            Reversal status flag.
        """
        return self._reverse_mode

    @reverse_mode.setter
    def reverse_mode(self, value: bool) -> None:
        self._reverse_mode = value

    def copy(self) -> "TranslationMap[NameType, SourceType, IdType]":
        """Make a copy of this ``TranslationMap``."""
        return copy(self)

    def __getitem__(self, name_or_source: Union[NameType, SourceType]) -> MagicDict[IdType]:
        return self.apply(name_or_source)

    def __len__(self) -> int:
        return len(self._names_and_sources)

    def __iter__(self) -> Iterator[Union[NameType, SourceType]]:
        return iter(self._names_and_sources)

    def __bool__(self) -> bool:
        return bool(self._format_appliers)

    def __repr__(self) -> str:
        sources = ", ".join(
            {f"'{formatter.source}': {len(formatter)} IDs" for formatter in self._format_appliers.values()}
        )
        return f"{tname(self)}({sources})"



"""Factory functions for translation classes."""
from pathlib import Path as _Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Generic as _Generic, Iterable, List, Optional, Tuple, Type, Union

from rics import misc
from rics._internal_support.types import PathLikeType
from rics.collections import dicts

from . import exceptions, fetching
from ._config_utils import ConfigMetadata as _ConfigMetadata
from ._load_toml import load_toml_file as _load_toml_file
from .mapping import HeuristicScore as _HeuristicScore, Mapper as _Mapper
from .types import IdType, NameType, SourceType

if TYPE_CHECKING:
    from . import Translator

FetcherFactory = Callable[[str, Dict[str, Any]], fetching.AbstractFetcher]
"""A callable which creates new ``AbstractFetcher`` instances from a dict config.

Config format is described in :ref:`translator-config-fetching`.

Args:
    clazz: Type of ``AbstractFetcher`` to create.
    config: Keyword arguments for the fetcher class.

Returns:
    A ``Fetcher`` instance.

Raises:
    exceptions.ConfigurationError: If `config` is invalid.
    TypeError: If `clazz` is not an AbstractFetcher subtype.
"""

MapperFactory = Callable[[Dict[str, Any], bool], Optional[_Mapper]]
"""A callable which creates new ``Mapper`` instances from a dict config.

Config format is described in :ref:`translator-config-mapping`.

If ``None`` is returned, a suitable default is used instead.

Args:
    config: Keyword arguments for the ``Mapper``.
    for_fetcher: Flag indicating that the ``Mapper`` returned will be used by an ``AbstractFetcher`` instance.

Returns:
    A ``Mapper`` instance or ``None``.

Raises:
    ConfigurationError: If `config` is invalid.
"""


def default_fetcher_factory(clazz: str, config: Dict[str, Any]) -> fetching.AbstractFetcher[SourceType, IdType]:
    """Create an ``AbstractFetcher`` from config."""
    fetcher_class = misc.get_by_full_name(clazz, default_module=fetching)
    fetcher = fetcher_class(**config)
    if not isinstance(fetcher, fetching.AbstractFetcher):  # pragma: no cover
        raise TypeError(f"Fetcher of type {fetcher} created from '{clazz}' is not an AbstractFetcher.")
    return fetcher


def default_mapper_factory(config: Dict[str, Any], for_fetcher: bool) -> Optional[_Mapper[Any, Any, Any]]:
    """Create a ``Mapper`` from config."""
    if "score_function" in config and isinstance(config["score_function"], dict):
        score_function = config.pop("score_function")

        if len(score_function) > 1:  # pragma: no cover
            raise exceptions.ConfigurationError(
                f"At most one score function may be specified, but got: {sorted(score_function)}"
            )

        score_function, score_function_kwargs = next(iter(score_function.items()))
        config["score_function"] = score_function
        config["score_function_kwargs"] = score_function_kwargs

    if "score_function_heuristics" in config:
        if "score_function" not in config:  # pragma: no cover
            section = "fetching" if for_fetcher else "translation"
            raise exceptions.ConfigurationError(
                f"Section [{section}.mapper.score_function_heuristics] requires an explicit score function."
            )

        heuristics = [
            (heuristic_config.pop("function"), heuristic_config)
            for heuristic_config in config.pop("score_function_heuristics")
        ]
        score_function = config["score_function"]

        if isinstance(score_function, _HeuristicScore):  # pragma: no cover
            for h, kwargs in heuristics:
                score_function.add_heuristic(h, kwargs)
        else:
            config["score_function"] = _HeuristicScore(score_function, heuristics)

    if "filter_functions" in config:
        config["filter_functions"] = [(f.pop("function"), f) for f in config.pop("filter_functions")]

    if "overrides" in config:  # pragma: no cover
        overrides = config.pop("overrides")
        shared, specific = _split_overrides(overrides)

        if specific and not for_fetcher:
            raise exceptions.ConfigurationError(
                "Context-sensitive overrides are not possible (or needed) for "
                f"Name-to-source mapping, but got {overrides=}."
            )

        config["overrides"] = dicts.InheritedKeysDict(specific, default=shared) if for_fetcher else shared

    return _Mapper(**config)


class TranslatorFactory(_Generic[NameType, SourceType, IdType]):
    """Create a ``Translator`` from TOML inputs."""

    FETCHER_FACTORY: FetcherFactory = default_fetcher_factory
    """A callable ``(clazz, config) -> AbstractFetcher``. Overwrite attribute to customize."""
    MAPPER_FACTORY: MapperFactory = default_mapper_factory
    """A callable ``(config, for_fetcher) -> Mapper``. Overwrite attribute to customize."""

    def __init__(
        self,
        file: PathLikeType,
        extra_fetchers: Iterable[PathLikeType],
        clazz: Union[str, Type["Translator[NameType, SourceType, IdType]"]] = None,
    ) -> None:
        self.file = str(file)
        self.extra_fetchers = list(map(str, extra_fetchers))
        self.clazz: Type[Translator[NameType, SourceType, IdType]] = self.resolve_class(clazz)
        self._metaconf = _read_metaconf(self.file)

    def create(self) -> "Translator[NameType, SourceType, IdType]":
        """Create a ``Translator`` from a TOML file."""
        config: Dict[str, Any] = self.load_toml_file(self.file)

        config_metadata = _ConfigMetadata.from_toml_paths(
            self.file,
            self.extra_fetchers,
            self.clazz,
        )

        _check_allowed_keys(["translator", "mapping", "fetching", "unknown_ids"], config, "<root>")

        translator_config = config.pop("translator", {})

        fetcher: fetching.Fetcher[SourceType, IdType] = self._handle_fetching(
            config.pop("fetching", {}), self.extra_fetchers, _cache_keys_from_config_metadata(config_metadata)
        )

        mapper = self._make_mapper("translator", translator_config)
        default_fmt_placeholders: Optional[dicts.InheritedKeysDict[SourceType, str, Any]]
        default_fmt, default_fmt_placeholders = _make_default_translations(**config.pop("unknown_ids", {}))

        ans = self.clazz(
            fetcher,
            mapper=mapper,
            default_fmt=default_fmt,
            default_fmt_placeholders=default_fmt_placeholders,
            **translator_config,
        )

        ans._config_metadata = config_metadata

        return ans

    def load_toml_file(self, path: str) -> Dict[str, Any]:
        """Read a TOML file from `path`."""
        config = dict(allow_interpolation=True)
        config.update(self._metaconf.get("env", {}))
        return _load_toml_file(path, **config)

    @classmethod
    def resolve_class(
        cls, clazz: Union[str, Type["Translator[NameType, SourceType, IdType]"]] = None
    ) -> Type["Translator[NameType, SourceType, IdType]"]:
        """Resolve desired ``Translator`` type."""
        if clazz is None:
            from . import Translator

            return Translator

        if isinstance(clazz, str):
            return misc.get_by_full_name(clazz)  # type: ignore[no-any-return]
        else:
            return clazz

    def _handle_fetching(
        self,
        config: Dict[str, Any],
        extra_fetchers: List[str],
        default_cache_keys: List[List[str]],
    ) -> fetching.Fetcher[SourceType, IdType]:
        multi_fetcher_kwargs = config.pop("MultiFetcher", {})

        fetchers: List[fetching.Fetcher[SourceType, IdType]] = []
        if config:
            fetchers.append(self._make_fetcher(default_cache_keys[0], **config))  # Add primary fetcher

        for i, file_fetcher_file in enumerate(extra_fetchers, start=1):
            fetchers.append(
                self._make_fetcher(default_cache_keys[i], **self.load_toml_file(file_fetcher_file)["fetching"])
            )

        if not fetchers:
            raise exceptions.ConfigurationError(
                "Section [fetching] is required when no pre-initialized AbstractFetcher is given."
            )

        return fetchers[0] if len(fetchers) == 1 else fetching.MultiFetcher(*fetchers, **multi_fetcher_kwargs)

    @classmethod
    def _make_mapper(cls, parent_section: str, config: Dict[str, Any]) -> Optional[_Mapper[Any, Any, Any]]:
        if "mapping" not in config:
            return None  # pragma: no cover

        config = config.pop("mapping")
        for_fetcher = parent_section.startswith("fetching")
        if for_fetcher:
            config = {**fetching.AbstractFetcher.default_mapper_kwargs(), **config}

        return TranslatorFactory.MAPPER_FACTORY(config, for_fetcher)

    @classmethod
    def _make_fetcher(cls, cache_keys: List[str], **config: Any) -> fetching.AbstractFetcher[SourceType, IdType]:
        mapper = cls._make_mapper("fetching", config) if "mapping" in config else None

        if len(config) == 0:  # pragma: no cover
            raise exceptions.ConfigurationError("Fetcher implementation section missing.")
        if len(config) > 1:  # pragma: no cover
            raise exceptions.ConfigurationError(
                f"Multiple fetcher implementations specified in the same file: {sorted(config)}"
            )

        clazz, kwargs = next(iter(config.items()))

        kwargs["cache_keys"] = kwargs.get("cache_keys", cache_keys)
        kwargs["mapper"] = mapper
        return TranslatorFactory.FETCHER_FACTORY(clazz, kwargs)


def _make_default_translations(
    **config: Any,
) -> Tuple[str, Optional[dicts.InheritedKeysDict[SourceType, str, Any]]]:  # pragma: no cover
    _check_allowed_keys(["fmt", "overrides"], config, toml_path="translator.unknown_ids")

    fmt = config.pop("fmt", None)
    if "overrides" in config:
        shared, specific = _split_overrides(config.pop("overrides"))
        return fmt, dicts.InheritedKeysDict(specific, default=shared)
    else:
        return fmt, None


def _check_allowed_keys(allowed: Iterable[str], actual: Iterable[str], toml_path: str) -> None:  # pragma: no cover
    bad_keys = set(actual).difference(allowed)
    if bad_keys:
        raise ValueError(f"Forbidden keys {sorted(bad_keys)} in [{toml_path}]-section.")


def _split_overrides(overrides: Any) -> Any:
    specific = {k: v for k, v in overrides.items() if isinstance(v, dict)}
    shared = {k: v for k, v in overrides.items() if k not in specific}
    return shared, specific


def _cache_keys_from_config_metadata(config_metadata: _ConfigMetadata) -> List[List[str]]:
    # Use the config filename and sha hash as the default keys
    return list(map(lambda t: [t[0].name, t[1]], (config_metadata.main,) + config_metadata.extra_fetchers))


def _read_metaconf(file: str) -> Dict[str, Any]:
    path = _Path(file).with_name("metaconf.toml")
    if path.exists():
        metaconf = _load_toml_file(path)
    else:
        return {}

    return metaconf



# noqa: D100

__title__ = "id-translation"
__description__ = "My personal little ML engineering library."
__version__ = "0.3.1"
__author__ = "Richard Sundqvist"
__copyright__ = "2022, Richard Sundqvist"



"""Utility module for parsing raw ``Format`` input strings."""

from dataclasses import dataclass as _dataclass
from string import Formatter as _Formatter
from typing import List, List as _List

OPTIONAL_BLOCK_START_DELIMITER = "["
OPTIONAL_BLOCK_END_DELIMITER = "]"

_formatter = _Formatter()
_hint = (
    f"Hint: Use double characters to escape '{OPTIONAL_BLOCK_START_DELIMITER}' and '{OPTIONAL_BLOCK_END_DELIMITER}', "
    f"e.g. '{OPTIONAL_BLOCK_START_DELIMITER * 2}' to render a single '{OPTIONAL_BLOCK_START_DELIMITER}'-character."
)


class MalformedOptionalBlockError(ValueError):
    """Error raised for improper optional blocks."""

    @staticmethod
    def get_marker_row(format_string: str, open_idx: int = -1, idx: int = -1) -> str:
        """Get a string of length equal to `format_string` which marks problem locations."""
        markers = [" "] * len(format_string)

        if idx != -1:
            markers[idx] = "^"
        if open_idx != -1:
            markers[open_idx] = "^"
        return "".join(markers)


class UnusedOptionalBlockError(MalformedOptionalBlockError):
    """Errors raised when optional blocks do not use any placeholders."""

    def __init__(self, format_string: str, open_idx: int, idx: int) -> None:
        problem_locations = MalformedOptionalBlockError.get_marker_row(format_string, open_idx, idx)
        super().__init__(
            f"""Malformed optional block. No {{placeholders}} found in range {open_idx, idx}).
'{format_string}'
 {problem_locations}
{_hint}""".strip()
        )


class BadDelimiterError(MalformedOptionalBlockError):
    """Errors raised due to mismatched delimiters."""

    def __init__(self, format_string: str, open_idx: int, idx: int) -> None:
        problem_locations = MalformedOptionalBlockError.get_marker_row(format_string, open_idx, idx)

        if idx == -1:
            super().__init__(
                f"""Malformed optional block. Optional block opened at i={open_idx} was never closed.
'{format_string}'
 {problem_locations}""".strip()
            )
        else:
            info = (
                "there is no block to close"
                if open_idx == -1
                else f"nested optional blocks are not supported (opened at {open_idx})"
            )

            super().__init__(
                f"""Malformed optional block. Got '{format_string[idx]}' at i={idx}, but {info}.
    '{format_string}'
     {problem_locations}
{_hint}""".strip()
            )


@_dataclass(frozen=True)
class Element:
    """Information about a single block in a ``Format`` specification."""

    part: str
    """String literal."""
    placeholders: _List[str]
    """Placeholder names in `part`, if any."""
    required: bool
    """Flag indicating whether the element may be excluded."""

    @property
    def positional_part(self) -> str:
        """Return a positional version of the `part` attribute."""
        if not self.placeholders:
            return self.part

        return self.part.format(**{p: "{}" for p in self.placeholders})

    @staticmethod
    def make(s: str, in_optional_block: bool) -> "Element":
        """Create an ``Element`` from an input string `s`.

        Args:
            s: Input data.
            in_optional_block: Flag indicating whether `s` was found inside an optional block.

        Returns:
            A new ``Element``.
        """
        parsed_block = s.replace("[[", "[").replace("]]", "]")
        placeholders = [x[1] for x in _formatter.parse(parsed_block) if x[1]]

        return Element(
            parsed_block,
            placeholders,
            not (placeholders and in_optional_block),
        )


def get_elements(fmt: str) -> List[Element]:
    """Split a format string into elements.

    Args:
        fmt: User input string.

    Returns:
        A list of parsed elements.

    Raises:
        BadDelimiterError: For unbalanced optional block delimitation characters.
        UnusedOptionalBlockError: If optional blocks are defined without placeholders.
    """
    if not fmt:
        return [Element("", [], required=True)]

    same_count = 1
    ans = []

    in_optional_block = fmt[0] == OPTIONAL_BLOCK_START_DELIMITER
    open_idx = 0 if in_optional_block else -1
    prev_idx = int(in_optional_block)

    for idx in range(int(in_optional_block), len(fmt)):
        char = fmt[idx]
        next_char = fmt[idx + 1] if idx + 1 < len(fmt) else None
        is_delimiter_char = char in (OPTIONAL_BLOCK_START_DELIMITER, OPTIONAL_BLOCK_END_DELIMITER)
        if next_char == char and is_delimiter_char:
            same_count += 1
        else:
            if same_count % 2 and is_delimiter_char:
                if char == OPTIONAL_BLOCK_START_DELIMITER:
                    if open_idx != -1:
                        raise BadDelimiterError(fmt, open_idx, idx)
                    open_idx = idx
                else:
                    if open_idx == -1:
                        raise BadDelimiterError(fmt, open_idx, idx)
                    open_idx = -1

                if prev_idx != idx:
                    element = Element.make(fmt[prev_idx:idx], in_optional_block)
                    if in_optional_block and not element.placeholders:
                        raise UnusedOptionalBlockError(fmt, prev_idx - 1, idx)
                    ans.append(element)
                in_optional_block = not in_optional_block
                prev_idx = idx + 1

            same_count = 1

    if prev_idx != len(fmt):
        ans.append(Element.make(fmt[prev_idx:], in_optional_block))

    if in_optional_block:
        raise BadDelimiterError(fmt, open_idx, -1)

    return ans



import logging
from typing import Any, Dict, Generic, Iterable, List, Optional, Tuple, Union

from rics.misc import get_by_full_name, tname

from . import filter_functions, heuristic_functions, score_functions
from .types import CandidateType, ContextType, HeuristicsTypes, ScoreFunction, ValueType

LOGGER = logging.getLogger(__package__).getChild("verbose").getChild("HeuristicScore")


class HeuristicScore(Generic[ValueType, CandidateType, ContextType]):
    """Callable wrapper for computing heuristic scores.

    Instances are callable. Signature is given by :attr:`~id_translation.mapping.types.ScoreFunction`.

    Short-circuiting:
        A mechanism for forced matching. Score is set to `+∞` for short-circuited candidates, and `-∞` for the rest.
        No further matching will be performed after this point, so ensure that all desired candidates are returned by
        chosen filters.

    Procedure:
        1. Trigger ``short-circuiting`` if there is an exact value-candidate match.
        2. All `heuristics` are applied and scores are computed.
        3. If no ``short-circuiting`` is triggered in step 2, yield max score for each candidate.

    Args:
        score_function: A :attr:`~id_translation.mapping.types.ScoreFunction` to wrap.
        heuristics: Iterable of heuristics or tuples ``(heuristic, kwargs)`` to apply to the ``(value, candidates)``
            inputs of `score_function`.

    Heuristic types:
        * An :const:`~id_translation.mapping.types.AliasFunction`, which accepts and returns a tuple
          (value, candidates) to be evaluated.
        * A :const:`~id_translation.mapping.types.FilterFunction`, which accepts a tuple (value, candidates) and
          returns a subset of `candidates`. If any candidates are returned, ``short-circuiting`` is triggered.

    Notes:
        * Heuristic function input order = application order.
        * You may add ``mutate=True`` to the heuristics kwargs to forward to the modifications made by that function.
    """

    def __init__(
        self,
        score_function: Union[str, ScoreFunction[ValueType, CandidateType, ContextType]],
        heuristics: Iterable[
            Union[
                Union[str, HeuristicsTypes[ValueType, CandidateType, ContextType]],
                Tuple[Union[str, HeuristicsTypes[ValueType, CandidateType, ContextType]], Dict[str, Any]],
            ]
        ],
    ) -> None:
        self._score: ScoreFunction[ValueType, CandidateType, ContextType] = (
            get_by_full_name(score_function, score_functions) if isinstance(score_function, str) else score_function
        )
        self._heuristics: List[Tuple[HeuristicsTypes[ValueType, CandidateType, ContextType], Dict[str, Any]]] = []

        for h in heuristics:
            func, kwargs = h if isinstance(h, tuple) else (h, {})
            self.add_heuristic(func, kwargs)

    @property
    def score_function(self) -> ScoreFunction[ValueType, CandidateType, ContextType]:
        """Return the underlying likeness score function."""
        return self._score

    def add_heuristic(
        self,
        heuristic: Union[str, HeuristicsTypes[ValueType, CandidateType, ContextType]],
        kwargs: Dict[str, Any] = None,
    ) -> None:
        """Add a new heuristic."""
        new_heuristic = (_resolve_heuristic(heuristic), kwargs or {})
        self._heuristics.append(new_heuristic)

    def __repr__(self) -> str:
        score_function = self._score
        heuristics = self._heuristics
        return f"{tname(self)}({score_function=}, {heuristics=})"

    def __call__(
        self, value: ValueType, candidates: Iterable[CandidateType], context: Optional[ContextType], **kwargs: Any
    ) -> Iterable[float]:
        """Apply `score_function` with heuristics and short-circuiting."""
        candidates = list(candidates)

        if value in candidates:
            yield from (float("inf") if c == value else -float("inf") for c in candidates)
            return

        base_score = list(self.score_function(value, candidates, context, **kwargs))  # Unmodified score
        best = list(base_score)

        log_aliases = heuristic_functions.VERBOSE and LOGGER.isEnabledFor(logging.DEBUG)

        positional_penalty = 0.0  # A small value that rewards alias functions based on their position.
        h_value = value
        h_candidates = list(candidates)
        for func, func_kwargs in self._heuristics:
            func_kwargs = func_kwargs.copy()
            mutate = func_kwargs.pop("mutate", False)
            res = func(h_value, h_candidates, context, **func_kwargs)
            if isinstance(res, tuple):  # Alias function -- res is a modified (value, candidates) tuple
                res_value, res_candidates = res[0], list(res[1])
                res_scores = list(self._score(res_value, res_candidates, context, **kwargs))
                for i, heuristic_score in enumerate(res_scores):
                    heuristic_score -= positional_penalty
                    best[i] = max(best[i], heuristic_score)

                if log_aliases:
                    mutating = "mutating" if mutate else "non-mutating"

                    res_value_repr = f"{res_value!r}" if (h_value != res_value) else "*"
                    res_candidates_repr = f"{res_candidates!r}" if (h_candidates != res_candidates) else "*"

                    res_score_repr = [round(s, 3) for s in res_scores]
                    LOGGER.debug(
                        f"Called {mutating} alias function {_stringify((func, func_kwargs))} in {context=}:\n    "
                        f"({h_value!r}, {h_candidates!r}) -> ({res_value_repr}, {res_candidates_repr})."
                        f"\n    Positional penalty={positional_penalty:.3f}. Scores before penalty: {res_score_repr}."
                    )
                else:
                    pass  # pragma: no cover

                if mutate:
                    h_value, h_candidates = res_value, res_candidates

                positional_penalty += 0.005
            elif res:  # Filter function
                if mutate:
                    raise TypeError(f"Filter function {_stringify((func, func_kwargs))} cannot use {mutate=}.")

                if heuristic_functions.VERBOSE and LOGGER.isEnabledFor(logging.DEBUG):
                    base_args = ", ".join([repr(h_value), repr(h_candidates), f"{context=}"])
                    extra_args = ", ".join(f"{k}={repr(v)}" for k, v, in func_kwargs.items())
                    info = f"{tname(func)}({', '.join([base_args, extra_args])})"
                    LOGGER.debug(f"Short-circuit {value=} -> candidates={repr(res)}, triggered by {info}.")
                yield from (float("inf") if c in res else -float("inf") for c in h_candidates)
                return  # Short-circuit

        if heuristic_functions.VERBOSE and LOGGER.isEnabledFor(logging.DEBUG):
            changes = [
                f"{repr(cand)}: {score:.2f} -> {heuristic_score:.2f} ({heuristic_score - score:+.2f})"
                for cand, score, heuristic_score in zip(candidates, base_score, best)
            ]
            LOGGER.debug(f"Heuristics scores for {value=}: [{', '.join(changes)}]")

        yield from best

    def __str__(self) -> str:
        score_function = tname(self.score_function, prefix_classname=True)
        return f"{tname(self)}([{' | '.join(map(_stringify, self._heuristics))}] -> {score_function})"


def _stringify(*args: Any) -> str:
    f, kwargs = args[0] if len(args) == 1 else args
    kwlist = (f"{k}={repr(v)}" for k, v in kwargs.items())
    return f"{tname(f)}({', '.join(kwlist)})"


def _resolve_heuristic(
    func_or_name: Union[str, HeuristicsTypes[ValueType, CandidateType, ContextType]]
) -> HeuristicsTypes[ValueType, CandidateType, ContextType]:
    if isinstance(func_or_name, str):
        for m in filter_functions, heuristic_functions:
            try:
                return get_by_full_name(func_or_name, m)  # type: ignore[no-any-return]
            except AttributeError:
                pass

        raise NameError(func_or_name)
    else:
        return func_or_name



"""Functions which return a likeness score.

See Also:
    The :class:`~.HeuristicScore` class.
"""
from typing import Iterable, Optional

from .types import CandidateType, ContextType, ValueType

VERBOSE: bool = False
"""If ``True`` enable optional DEBUG-level log messages on each score function invocation.

Notes:
    Not all functions have verbose messages.
"""


def modified_hamming(
    name: str,
    candidates: Iterable[str],
    context: Optional[ContextType],
    add_length_ratio_term: bool = True,
    positional_penalty: float = 0.001,
) -> Iterable[float]:
    """Compute hamming distance modified by length ratio, from the back. Score range is ``[0, 1]``.

    Keyword Args:
        add_length_ratio_term: If ``True``, score is divided by ``abs(len(name) - len(candidate))``.
        positional_penalty: A penalty applied to prefer earlier `candidates`, according to the formulare
            ``penalty = index(candidate) * positional_penalty)``.

    Examples:
        >>> from id_translation.mapping.score_functions import modified_hamming
        >>> print(list(modified_hamming('aa', ['aa', 'a', 'ab', 'aa'], context=None)))
        [1.0, 0.499, 0.498, 0.997]
        >>> print(list(modified_hamming('aa', ['aa', 'a', 'ab', 'aa'], context=None, positional_penalty=0)))
        [1.0, 0.5, 0.5, 1.0]
        >>> print(list(modified_hamming('face', ['face', 'FAce', 'race', 'place'], context=None)))
        [1.0, 0.499, 0.748, 0.372]
    """

    def _apply(candidate: str) -> float:
        sz = min(len(candidate), len(name))
        same = sum([name[i] == candidate[i] for i in range(-sz, 0)])

        ratio = (1 / (1 + abs(len(candidate) - len(name)))) if add_length_ratio_term else 1
        normalized_hamming = same / sz

        return ratio * normalized_hamming

    yield from (s - i * positional_penalty for i, s in enumerate(map(_apply, candidates)))


def equality(value: ValueType, candidates: Iterable[CandidateType], context: Optional[ContextType]) -> Iterable[float]:
    """Return 1.0 if ``k == c_i``, 0.0 otherwise.

    Examples:
        >>> from id_translation.mapping.score_functions import equality
        >>> print(list(equality('a', 'aAb', context=None)))
        [1.0, 0.0, 0.0]
    """
    yield from map(float, (value == c for c in candidates))


def disabled(value: ValueType, candidates: Iterable[CandidateType], context: Optional[ContextType]) -> Iterable[float]:
    """Special value to indicate that scoring logic has been disabled.

    This is a workaround to allow users to indicate that the scoring logic is disabled, and that overrides should be
    used instead. The ``disabled``-function has no special meaning to the mapper, and will be called as any other
    scoring function. This in turn will immediately raise a ``ScoringDisabledError``.

    Raises:
        ScoringDisabledError: Always.
    """
    from .exceptions import ScoringDisabledError

    raise ScoringDisabledError(value, candidates, context)



from typing import Any, Dict, List, Sequence

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType
from . import DataStructureIO
from .exceptions import NotInplaceTranslatableError


class SingleValueIO(DataStructureIO):
    """Implementation for non-iterables. And strings."""

    @staticmethod
    def handles_type(arg: Any) -> bool:
        return isinstance(arg, (int, str))

    @staticmethod
    def extract(translatable: IdType, names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        if len(names) != 1:  # pragma: no cover
            raise ValueError("Length of names must be one.")

        return {names[0]: (translatable,)}

    @staticmethod
    def insert(
        translatable: IdType, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> str:
        if not copy:  # pragma: no cover
            raise NotInplaceTranslatableError(translatable)

        return tmap[names[0]][translatable]



import logging
import pickle  # noqa: 403
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import Any, Dict, Generic, List, Optional, Union

import pandas as pd

from id_translation._base_metadata import BaseMetadata
from id_translation.types import HasSources, IdType, SourceType

from ..offline.types import PlaceholderTranslations, SourcePlaceholderTranslations

BASE_LOGGER = logging.getLogger(__package__).getChild("CacheMetadata")


@dataclass(eq=False)
class CacheMetadata(BaseMetadata, Generic[SourceType, IdType], HasSources[SourceType]):
    """Metadata pertaining to fetcher caching logic.

    Args:
        cache_keys: Hierarchical identifiers for the cache. The first key is used to determine storage location, while
            the rest are used to detect configuration changes (which invalidate the cache). A typical key would be
            ``[config-file-name, config-file-sha]``.
        placeholders: A Source-to-placeholder dict.
        **kwargs: Forwarded to base classes.
    """

    def __init__(
        self,
        cache_keys: List[str],
        placeholders: Dict[SourceType, List[str]],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if not cache_keys:  # pragma: no cover
            raise ValueError("Must specify at least one cache key.")
        self._cache_keys = cache_keys
        self._placeholders = placeholders

    @property
    def cache_keys(self) -> List[str]:
        """Yields hierarchical cache keys for this metadata."""
        return self._cache_keys.copy()

    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        return self._placeholders

    def _is_equivalent(self, other: "BaseMetadata") -> str:
        assert isinstance(other, CacheMetadata)  # noqa: S101

        if self._cache_keys != other._cache_keys:
            return f"Expected cache_keys={self._cache_keys} (requested) but got {other._cache_keys}"

        if list(self.placeholders) != list(other.placeholders):
            return f"Expected sources={list(self.placeholders)} (requested) but got {list(other.placeholders)}"

        for source, expected in self.placeholders.items():
            if sorted(expected) != sorted(other.placeholders[source]):
                return f"For {source=}, expected placeholders={expected} but got {other.placeholders[source]}"

        return ""

    @property
    def _logger(self) -> logging.Logger:
        return BASE_LOGGER.getChild(self.cache_keys[0])

    def _log_reject(self, msg: str) -> None:
        self._logger.debug(msg.format(slug="Fetch new translations"))  # noqa: G001

    def _log_accept(self, msg: str) -> None:
        self._logger.debug(msg.format(kind="translations"))  # noqa: G001

    def _serialize(self, to_json: Dict[str, Any]) -> Dict[str, Any]:
        ans = to_json.copy()
        to_json.clear()
        return ans

    @classmethod
    def _deserialize(cls, from_json: Dict[str, Any]) -> Dict[str, Any]:
        return dict(cache_keys=from_json.pop("_cache_keys"), placeholders=from_json.pop("_placeholders"))


class CacheAccess(Generic[SourceType, IdType]):
    """Utility class for managing the FETCH_ALL-cache.

    Args:
        max_age: Cache timeout.
        metadata: Metadata object used to determine cache validity.
    """

    @classmethod
    def base_cache_dir_for_all_fetchers(cls) -> Path:
        """Top-level cache dir for all fetchers managed by any ``CacheAccess``-instance."""
        return Path.home().absolute().joinpath(".cache/id-translation/cached-fetcher-data/")

    @classmethod
    def clear_all_cache_data(cls) -> None:
        """Remove the entire cache directory tree for ALL instances."""
        import shutil

        cache_dir = cls.base_cache_dir_for_all_fetchers()
        print(f"Deleting the common cache directory tree at:\n    '{cache_dir}'")
        shutil.rmtree(cache_dir)

    def __init__(
        self,
        max_age: Union[str, pd.Timedelta, timedelta],
        metadata: CacheMetadata[SourceType, IdType],
    ) -> None:
        pd_timedelta = pd.Timedelta(max_age)
        if pd_timedelta < pd.Timedelta(0):
            raise ValueError("fetch_all_cache_max_age must be non-negative")  # pragma: no cover
        self._max_age = pd_timedelta.to_pytimedelta()
        self._metadata: CacheMetadata[SourceType, IdType] = metadata
        top_level_key = next(iter(self._metadata.cache_keys))
        self._base_dir = self.base_cache_dir_for_all_fetchers().joinpath(top_level_key)
        self._logger = logging.getLogger(__package__).getChild("CacheAccess").getChild(top_level_key)

    @property
    def cache_dir(self) -> Path:
        """Get the cache directory used by this ``CacheAccess``.

        Created from :meth:`base_cache_dir_for_all_fetchers` and the first value of :attr:`CacheMetadata.cache_keys`.

        Returns:
            Cache dir for a single fetcher.
        """
        return self._base_dir

    @property
    def metadata_path(self) -> Path:
        return self._base_dir.joinpath("metadata.json")

    @property
    def data_path(self) -> Path:
        return self._base_dir.joinpath("data.pkl")

    def read_cache(self, source: SourceType) -> Optional[PlaceholderTranslations[SourceType]]:
        """Read cached translation data for `source`."""
        if not self._metadata.use_cached(self.metadata_path, self._max_age):
            return None

        with self.data_path.open("rb") as f:
            try:
                spt = pickle.load(f)  # noqa: S301
                ans = spt[source]
                if not isinstance(ans, PlaceholderTranslations):  # pragma: no cover
                    reason = f"Got {type(ans).__name__} but expected {PlaceholderTranslations.__name__}."
                    self.clear(reason)
                    raise TypeError(reason)

                self._logger.debug(f"Returning cached data for {source=}", extra=dict(source=source))
                return ans
            except (TypeError, KeyError, pickle.UnpicklingError) as e:
                # These errors indicate a likely corrupted cache. Any other is probably the users' fault.
                f.close()
                self.clear(f"Got error {e} while deserializing", log_level=logging.ERROR, exc_info=True)
        return None

    def write_cache(self, data: SourcePlaceholderTranslations[SourceType]) -> None:
        """Overwrite the current cache (if it exists) with new and update metadata."""
        if self.metadata_path.exists():
            self._logger.debug(f"Overwriting existing cache data in '{self.cache_dir}'.")

        self.cache_dir.mkdir(parents=True, exist_ok=True)

        self.metadata_path.write_text(self._metadata.to_json())
        with self.data_path.open("wb") as f:
            pickle.dump(data, f)

    def clear(self, reason: str, log_level: int = logging.DEBUG, *, exc_info: bool = False) -> None:
        """Remove cached data for the current instance."""
        reason = ": " + reason

        deleted = []

        if self.data_path.exists():
            self.data_path.unlink(missing_ok=True)
            deleted.append(self.data_path)

        if self.metadata_path.exists():
            self.metadata_path.unlink(missing_ok=True)
            deleted.append(self.metadata_path)

        if self._logger.isEnabledFor(log_level):
            delete_info = "\n".join(map("    - {}".format, deleted)) if deleted else "    There is nothing to delete."
            self._logger.log(
                log_level,
                f"Clear cache: {reason}. The following files have been deleted:\n{delete_info}",
                exc_info=exc_info,
            )



"""Functions and classes used by the ``Mapper`` for handling score matrices.

.. warning::

   This module is considered an implementation detail, and may change without notice.
"""

import logging
import warnings
from collections import defaultdict as _defaultdict
from contextlib import contextmanager as _contextmanager
from dataclasses import dataclass as _dataclass
from typing import Dict, Generic as _Generic, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd

from ._cardinality import Cardinality as _Cardinality
from ._directional_mapping import DirectionalMapping as _DirectionalMapping
from .exceptions import AmbiguousScoreError as _AmbiguousScoreError
from .types import CandidateType, ValueType

warnings.warn(
    "This module is considered an implementation detail, and may change without notice.", UserWarning, stacklevel=2
)
_MATCH_SCORES_LOGGER = logging.getLogger(__package__).getChild("MatchScores")


@_contextmanager
def enable_verbose_debug_messages():  # type: ignore  # noqa
    """Temporarily enable verbose DEBUG-level logger messages.

    Returns a context manager. Calling the function without the ``with`` statement does nothing.

    >>> from id_translation.mapping import Mapper, support
    >>> with support.enable_verbose_debug_messages():
    ...     Mapper().apply("ab", candidates="abc")
    """  # noqa: DAR301
    from id_translation.mapping import VERBOSE_LOGGER, _mapper, filter_functions, heuristic_functions, score_functions

    before = filter_functions.VERBOSE, heuristic_functions.VERBOSE, score_functions.VERBOSE, VERBOSE_LOGGER.disabled
    enable = (True, True, True, False)
    try:
        filter_functions.VERBOSE, heuristic_functions.VERBOSE, score_functions.VERBOSE, VERBOSE_LOGGER.disabled = enable
        _mapper.FORCE_VERBOSE = True
        yield
    finally:
        filter_functions.VERBOSE, heuristic_functions.VERBOSE, score_functions.VERBOSE, VERBOSE_LOGGER.disabled = before
        _mapper.FORCE_VERBOSE = False


class MatchScores:
    """High-level selection operations.

    Args:
        scores: A score matrix, where ``scores.index`` are values and ``score.columns`` are treated as the candidates.
        min_score: Minimum score to consider make a `value -> candidate` match.
        logger: Explicit ``Logger`` instance to use.
    """

    def __init__(self, scores: pd.DataFrame, min_score: float, logger: logging.Logger = None) -> None:
        self._min_score = min_score
        self._matrix = scores
        self._logger = _MATCH_SCORES_LOGGER if logger is None else logger

    @property
    def logger(self) -> logging.Logger:
        """Return the ``Logger`` that is used by this instance."""
        return self._logger

    def to_directional_mapping(self, cardinality: _Cardinality = None) -> _DirectionalMapping[ValueType, CandidateType]:
        """Create a ``DirectionalMapping`` with a given target ``Cardinality``.

        Args:
            cardinality: Explicit cardinality to set, see :attr:`~.DirectionalMapping.cardinality`. If ``None``, use the
                actual cardinality when selecting all matches with scores :attr:`get_above` the minimum.

        Returns:
            A ``DirectionalMapping``.
        """
        matches: List[MatchScores.Record[ValueType, CandidateType]]
        rejections: List[MatchScores.Reject[ValueType, CandidateType]]
        matches, rejections = self._match(cardinality)

        left_to_right = _defaultdict(list)
        for record in list(matches):
            supersedes: List[MatchScores.Reject[ValueType, CandidateType]] = []
            if self.logger.isEnabledFor(logging.DEBUG) and rejections:
                for rr in rejections:
                    if record in (rr.superseding_value, rr.superseding_candidate):
                        supersedes.append(rr)

            if self.logger.isEnabledFor(logging.INFO):
                reason = "(short-circuit or override)" if record.score == np.inf else f">= {self._min_score}"
                self.logger.debug(f"Accepted: {record} {reason}.")

            if supersedes:
                s = "\n".join("    " + rr.explain(self._min_score) for rr in supersedes)
                self.logger.debug(f"This match supersedes {len(supersedes)} other matches:\n{s}")

            left_to_right[record.value].append(record.candidate)

        if rejections and self.logger.isEnabledFor(logging.DEBUG):
            unmapped_values = set(self._matrix.index.difference(left_to_right))
            for value in unmapped_values:
                lst = []
                for rr in filter(lambda r: r.record.value == value, rejections):  # noqa: B023
                    lst.append(f"    {rr.explain(self._min_score, full=True)}")
                value_reasons = "\n".join(lst)
                self.logger.debug(f"Could not map {value=}:\n{value_reasons}")

        return _DirectionalMapping(
            cardinality=cardinality,
            left_to_right={
                value: tuple(left_to_right[value]) for value in self._matrix.index if value in left_to_right
            },
            _verify=False,
        )

    def _match(
        self, cardinality: _Cardinality = None
    ) -> Tuple[List["MatchScores.Record[ValueType, CandidateType]"], List["Reject[ValueType, CandidateType]"]]:
        rejections: Optional[List[MatchScores.Reject[ValueType, CandidateType]]] = None
        records: List["MatchScores.Record[ValueType, CandidateType]"] = self.get_above()

        if self.logger.isEnabledFor(logging.DEBUG):
            rejections = []
            records.extend(self.get_below())

        if cardinality is _Cardinality.OneToOne:
            matches = self._select_one_to_one(records, rejections)
        elif cardinality is _Cardinality.OneToMany:
            matches = self._select_one_to_many(records, rejections)
        elif cardinality is _Cardinality.ManyToOne:
            matches = self._select_many_to_one(records, rejections)
        else:
            matches = self._select_many_to_many(records, rejections)

        return list(matches), rejections or []

    def _get_sorted(self) -> pd.Series:
        sorted_scores: pd.Series = self._matrix.stack()
        sorted_scores.sort_values(ascending=False, inplace=True)
        return sorted_scores

    def get_above(self) -> List["MatchScores.Record[ValueType, CandidateType]"]:
        """Get all records with scores `above` the threshold."""
        s = self._get_sorted()
        return self._from_series(s[s >= self._min_score])

    def get_below(self) -> List["MatchScores.Record[ValueType, CandidateType]"]:
        """Get all records with scores `below` the threshold."""
        s = self._get_sorted()
        return self._from_series(s[s < self._min_score])

    @_dataclass(frozen=True)
    class Record(_Generic[ValueType, CandidateType]):
        """Data concerning a match."""

        value: ValueType
        """A hashable value."""
        candidate: CandidateType
        """A hashable candidate."""
        score: float
        """Likeness score computed by some scoring function."""

        def __str__(self) -> str:
            return f"{repr(self.value)} -> '{self.candidate}'; score={self.score:.3f}"

    @classmethod
    def _from_series(cls, s: pd.Series) -> List[Record[ValueType, CandidateType]]:
        return [MatchScores.Record(value, candidate, score) for (value, candidate), score in s.items()]

    @_dataclass(frozen=True)
    class Reject(_Generic[ValueType, CandidateType]):
        """Data concerning the rejection of a match."""

        record: "MatchScores.Record[ValueType, CandidateType]"
        superseding_value: Optional["MatchScores.Record[ValueType, CandidateType]"] = None
        superseding_candidate: Optional["MatchScores.Record[ValueType, CandidateType]"] = None

        def explain(self, min_score: float, full: bool = False) -> str:
            """Create a string which explains the rejection.

            Args:
                min_score: Minimum score to accept a match.
                full: If ``True`` show full information about superseding matches.

            Returns:
                An explanatory string.
            """
            if self.record.score == -np.inf:
                if self.superseding_value and self.superseding_value.score == np.inf:
                    extra = f": {self.superseding_value}" if full else ""
                    why = f" (superseded by short-circuit or override{extra})"
                elif self.superseding_candidate and self.superseding_candidate.score == np.inf:
                    extra = f": {self.superseding_candidate}" if full else ""
                    why = f" (superseded by short-circuit or override{extra}"
                else:
                    why = " (filtered)"
            elif self.record.score < min_score:
                why = f" < {min_score} (below threshold)"
            else:
                ands = []
                if self.superseding_value:
                    extra = f": {self.superseding_value}" if full else ""
                    ands.append(f"value={repr(self.superseding_value.value)}{extra}")
                if self.superseding_candidate:
                    extra = f": {self.superseding_candidate}" if full else ""
                    ands.append(f"candidate={repr(self.superseding_candidate.candidate)}{extra}")
                why = f" (superseded on {' and '.join(ands)})"

            return f"{self.record}{why}."

    def _raise_if_ambiguous(
        self,
        record: Record,  # type: ignore[type-arg]
        matches: Dict,  # type: ignore[type-arg]
        kind: str,
        cardinality: _Cardinality,
    ) -> None:
        if record.score == np.inf:
            # Overrides are allowed to be infinite; the first one will be chosen. It's up to the user to manage them.
            return

        key = record.value if kind == "value" else record.candidate
        if key not in matches:
            return

        old_match = matches[key]
        if record.score == old_match.score:
            raise _AmbiguousScoreError(
                kind=kind,
                key=key,
                match0=record,
                match1=old_match,
                cardinality=cardinality.name,
                scores=self._matrix.to_string(),
            )

    def _select_one_to_one(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: List[Reject[ValueType, CandidateType]] = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        mvs: Dict[ValueType, MatchScores.Record[ValueType, CandidateType]] = {}
        mcs: Dict[CandidateType, MatchScores.Record[ValueType, CandidateType]] = {}

        for record in records:
            self._raise_if_ambiguous(record, mcs, "candidate", _Cardinality.OneToOne)
            self._raise_if_ambiguous(record, mvs, "value", _Cardinality.OneToOne)

            if record.score < self._min_score or record.value in mvs or record.candidate in mcs:
                if rejections is not None:
                    rejections.append(
                        MatchScores.Reject(
                            record,
                            superseding_value=mvs.get(record.value),
                            superseding_candidate=mcs.get(record.candidate),
                        )
                    )
                continue
            mvs[record.value] = record
            mcs[record.candidate] = record
            yield record

    def _select_one_to_many(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: List[Reject[ValueType, CandidateType]] = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        mcs: Dict[CandidateType, MatchScores.Record[ValueType, CandidateType]] = {}

        for record in records:
            self._raise_if_ambiguous(record, mcs, "candidate", _Cardinality.OneToMany)

            if record.score < self._min_score or record.candidate in mcs:
                if rejections is not None:
                    rejections.append(MatchScores.Reject(record, superseding_candidate=mcs.get(record.candidate)))
                continue
            mcs[record.candidate] = record
            yield record

    def _select_many_to_one(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: List[Reject[ValueType, CandidateType]] = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        mvs: Dict[ValueType, MatchScores.Record[ValueType, CandidateType]] = {}

        for record in records:
            self._raise_if_ambiguous(record, mvs, "value", cardinality=_Cardinality.ManyToOne)

            if record.score < self._min_score or record.value in mvs:
                if rejections is not None:
                    rejections.append(MatchScores.Reject(record, superseding_value=mvs.get(record.value)))
                continue
            mvs[record.value] = record
            yield record

    def _select_many_to_many(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: List[Reject[ValueType, CandidateType]] = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        for record in records:
            if record.score < self._min_score:
                if rejections is not None:
                    rejections.append(MatchScores.Reject(record))
                continue
            yield record



from typing import Any, Iterator, Mapping, Optional

from ..types import IdType
from .types import TranslatedIds


class MagicDict(Mapping[IdType, str]):
    """Immutable mapping for translated IDs.

    If `default_value` is given, it is used as the default answer for any calls to `__getitem__` where the key is
    not in `translated_ids`.

    Args:
        real_translations: A dict holding real translations.
        default_value: A string with exactly one or zero placeholders.
    """

    def __init__(
        self,
        real_translations: TranslatedIds[IdType],
        default_value: str = None,
    ) -> None:
        self._real: TranslatedIds[IdType] = real_translations
        self._default = default_value

    @property
    def default_value(self) -> Optional[str]:
        """Return the default string value to return for unknown keys, if any."""
        return self._default  # pragma: no cover

    def __getitem__(self, key: IdType) -> str:
        if key in self._real or self._default is None:
            return self._real[key]

        return self._default.format(key)

    def __contains__(self, idx: Any) -> bool:
        return self._default is not None or idx in self._real  # pragma: no cover

    def __len__(self) -> int:
        return len(self._real)  # pragma: no cover

    def __iter__(self) -> Iterator[IdType]:
        return iter(self._real)  # pragma: no cover

    def __repr__(self) -> str:
        return repr(self._real)  # pragma: no cover



"""Types used for offline translation."""
import dataclasses
from typing import TYPE_CHECKING, Any, Dict, Generic as _Generic, Sequence, Tuple, Union

import pandas as pd

from ..types import ID, IdType, SourceType

if TYPE_CHECKING:
    from ._format import Format

FormatType = Union[str, "Format"]

PlaceholdersTuple = Tuple[str, ...]
TranslatedIds = Dict[IdType, str]  # {id: translation}


@dataclasses.dataclass
class PlaceholderTranslations(_Generic[SourceType]):
    """Matrix of ID translation components returned by fetchers."""

    MakeTypes = Union["PlaceholderTranslations", pd.DataFrame, Dict[str, Sequence[Any]]]

    source: SourceType
    """Source from which translations were retrieved."""
    placeholders: PlaceholdersTuple
    """Names of placeholders in the order in which they appear in `records`."""
    records: Sequence[Sequence[Any]]
    """Matrix of shape `N x M` where `N` is the number of IDs returned and `M` is the length of `placeholders`."""
    id_pos: int = -1
    """Position if the the ID placeholder in `placeholders`."""

    @classmethod
    def make(cls, source: SourceType, data: MakeTypes) -> "PlaceholderTranslations[SourceType]":
        """Try to make in instance from arbitrary input data.

        Args:
            source: Source label for the translations.
            data: Some data to convert to a PlaceholderTranslations instance.

        Returns:
            A new PlaceholderTranslations instance.

        Raises:
            TypeError: If `data` cannot be converted.
        """
        if isinstance(data, PlaceholderTranslations):  # pragma: no cover
            return data

        if isinstance(data, pd.DataFrame):
            return cls.from_dataframe(source, data)
        if isinstance(data, dict):
            return cls.from_dict(source, data)

        raise TypeError(data)  # pragma: no cover

    def to_dict(self, max_rows: int = 0) -> Dict[str, Sequence[Any]]:
        """Create a dict representation of the translations."""
        records = self.records[:max_rows] if max_rows else self.records
        return {placeholder: [row[i] for row in records] for i, placeholder in enumerate(self.placeholders)}

    @staticmethod
    def to_dicts(
        source_translations: "SourcePlaceholderTranslations[SourceType]",
        max_rows: int = 0,
    ) -> Dict[SourceType, Dict[str, Sequence[Any]]]:
        """Create a nested dict representation of the translations."""
        return {source: translations.to_dict(max_rows) for source, translations in source_translations.items()}

    @classmethod
    def from_dataframe(cls, source: SourceType, data: pd.DataFrame) -> "PlaceholderTranslations[SourceType]":
        """Create instance from a pandas DataFrame."""
        return cls(
            source,
            placeholders=tuple(data),
            records=data.values.tolist(),
            id_pos=data.columns.get_loc(ID) if ID in data else -1,
        )

    @classmethod
    def from_dict(cls, source: SourceType, data: Dict[str, Sequence[Any]]) -> "PlaceholderTranslations[SourceType]":
        """Create instance from a dict."""
        return cls.from_dataframe(source, pd.DataFrame.from_dict(data))


SourcePlaceholderTranslations = Dict[SourceType, PlaceholderTranslations[SourceType]]



from enum import Enum
from typing import Tuple, Union


class Cardinality(Enum):
    """Enumeration type for cardinality relationships.

    Cardinalities are comparable using numerical operators, and can be thought of as comparing "preciseness". The less
    ambiguity there is for a given cardinality, the smaller it is in comparison to the others. The hierarchy is given by
    ``1:1 < 1:N = N:1 < M:N``. Note that ``1:N`` and ``N:1`` are considered equally precise.

    Examples:
        Comparing cardinalities

        >>> from id_translation.mapping import Cardinality
        >>> Cardinality.ManyToOne
        <Cardinality.ManyToOne: 'N:1'>
        >>> Cardinality.OneToOne
        <Cardinality.OneToOne: '1:1'>
        >>> Cardinality.ManyToOne < Cardinality.OneToOne
        False
    """

    _ignore_ = ["ParseType"]

    ParseType = Union[str, "Cardinality"]  # Type checking
    """Types that may be interpreted as a ``Cardinality``."""

    OneToOne = "1:1"
    """One-to-one relationship."""
    OneToMany = "1:N"
    """One-to-many relationship."""
    ManyToOne = "N:1"
    """Many-to-one relationship."""
    ManyToMany = "M:N"
    """Many-to-many relationship."""

    @property
    def many_left(self) -> bool:
        """Many-relationship on the left, ``True`` for ``N:1`` and ``M:N``."""
        return self == Cardinality.ManyToMany or self == Cardinality.ManyToOne

    @property
    def many_right(self) -> bool:
        """Many-relationship on the right, ``True`` for ``1:N`` and ``M:N``."""
        return self == Cardinality.ManyToMany or self == Cardinality.OneToMany

    @property
    def one_left(self) -> bool:
        """One-relationship on the left, ``True`` for ``1:1`` and ``1:N``."""
        return not self.many_left

    @property
    def one_right(self) -> bool:
        """One-relationship on the right, ``True`` for ``1:1`` and ``N:1``."""
        return not self.many_right

    @property
    def inverse(self) -> "Cardinality":
        """Inverse cardinality. For symmetric cardinalities, ``self.inverse == self``.

        Returns:
            Inverse cardinality.

        See Also:
            :attr:`symmetric`
        """
        if self == Cardinality.OneToMany:
            return Cardinality.ManyToOne
        if self == Cardinality.ManyToOne:
            return Cardinality.OneToMany

        return self

    @property
    def symmetric(self) -> bool:
        """Symmetry flag. For symmetric cardinalities, ``self.inverse == self``.

        Returns:
            Symmetry flag.

        See Also:
            :attr:`inverse`
        """
        return self == Cardinality.OneToOne or self == Cardinality.ManyToMany

    def __ge__(self, other: "Cardinality") -> bool:
        """Equivalent to :meth:`set.issuperset`."""
        return _is_superset(self, other)

    def __lt__(self, other: "Cardinality") -> bool:
        return not self >= other

    @classmethod
    def from_counts(cls, left_count: int, right_count: int) -> "Cardinality":
        """Derive a `Cardinality` from counts.

        Args:
            left_count: Number of elements on the left-hand side.
            right_count: Number of elements on the right-hand side.

        Returns:
            A :class:`Cardinality`.

        Raises:
            ValueError: For counts < 1.
        """
        return _from_counts(left_count, right_count)

    @classmethod
    def parse(cls, arg: ParseType, strict: bool = False) -> "Cardinality":
        """Convert to cardinality.

        Args:
            arg: Argument to parse.
            strict: If ``True``, `arg` must match exactly when it is given as a string.

        Returns:
            A :class:`Cardinality`.

        Raises:
            ValueError: If the argument could not be converted.
        """
        return arg if isinstance(arg, Cardinality) else _from_generous_string(arg, strict)


Cardinality.ParseType = Union[str, Cardinality]  # type: ignore[assignment]


def _parsing_failure_message(arg: str, strict: bool) -> str:
    options = tuple([c.value for c in Cardinality])
    alternatively = tuple([c.name for c in Cardinality])
    strict_hint = "."
    if strict:
        try:
            strict = False
            Cardinality.parse(arg, strict=strict)
            strict_hint = f". Hint: set {strict=} to allow this input."
        except ValueError:
            pass
    return f"Could not convert {arg=} to Cardinality{strict_hint} Correct input {options=} or {repr(alternatively)}"


_MATRIX = (
    (Cardinality.ManyToMany, Cardinality.ManyToOne),
    (Cardinality.OneToMany, Cardinality.OneToOne),
)


def _is_superset(c0: Cardinality, c1: Cardinality) -> bool:
    if c0 == c1:
        return True

    c0_i, c0_j = _pos(c0)
    c1_i, c1_j = _pos(c1)
    return c0_i <= c1_i and c0_j <= c1_j


def _pos(cardinality: Cardinality) -> Tuple[int, int]:
    for i in range(2):
        for j in range(2):
            if _MATRIX[i][j] == cardinality:
                return i, j
    raise AssertionError("This should be impossible.")


def _from_counts(left_count: int, right_count: int) -> Cardinality:
    if left_count < 1:
        raise ValueError(f"{left_count=} < 1")
    if right_count < 1:
        raise ValueError(f"{right_count=} < 1")

    one_left = left_count == 1
    one_right = right_count == 1

    return _MATRIX[int(one_left)][int(one_right)]


def _from_generous_string(s: str, strict: bool) -> Cardinality:
    if not strict:
        s = s.strip().upper().replace("-", ":", 1).replace("*", "N", 2)
        if s == "N:N":
            s = "M:N"
    for c in Cardinality:
        if s == c.value or s == c.name:
            return c
    raise ValueError(_parsing_failure_message(s, strict))



import logging
import warnings
from time import perf_counter
from typing import Any, Dict, Generic, Iterable, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
from rics.action_level import ActionLevel
from rics.collections.dicts import InheritedKeysDict
from rics.misc import get_by_full_name, tname
from rics.performance import format_perf_counter

from . import exceptions, filter_functions as mf, score_functions as sf
from ._cardinality import Cardinality
from ._directional_mapping import DirectionalMapping
from .exceptions import MappingError, MappingWarning, UserMappingError, UserMappingWarning
from .types import CandidateType, ContextType, FilterFunction, ScoreFunction, UserOverrideFunction, ValueType

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=UserWarning)
    from .support import MatchScores, enable_verbose_debug_messages

FORCE_VERBOSE: bool = False  # Magic variable used by the verbose context


class Mapper(Generic[ValueType, CandidateType, ContextType]):
    """Optimal value-candidate matching.

    For an introduction to mapping, see the :ref:`mapping-primer` page.

    Args:
        score_function: A callable which accepts a value `k` and an ordered collection of candidates `c`, returning a
            score ``s_i`` for each candidate `c_i` in `c`. Default: ``s_i = float(k == c_i)``. Higher=better match.
        score_function_kwargs: Keyword arguments for `score_function`.
        filter_functions: Function-kwargs pairs of filters to apply before scoring.
        min_score: Minimum score `s_i`, as given by ``score(k, c_i)``, to consider `k` a match for `c_i`.
        overrides: If a dict, assumed to be 1:1 mappings (`value` to `candidate`) which override the scoring logic. If
            :class:`rics.collections.dicts.InheritedKeysDict`, the context passed to :meth:`apply` is used to retrieve
            specific overrides.
        unmapped_values_action: Action to take if mapping fails for any values.
        unknown_user_override_action: Action to take if a :attr:`id_translation.mapping.types.UserOverrideFunction`
            returns an unknown candidate. Unknown candidates, i.e. candidates not in the input `candidates` collection,
            will not be used unless `'ignore'` is chosen. As such, `'ignore'` should rather be interpreted as `'allow'`.
        cardinality: Desired cardinality for mapped values. Derive for each matching if ``None``.
        verbose_logging: If ``True``, enable verbose logging for the :meth:`apply` function. Has no effect when the log
            level is above ``logging.DEBUG``.
    """

    def __init__(
        self,
        score_function: Union[str, ScoreFunction[ValueType, CandidateType, ContextType]] = "equality",
        score_function_kwargs: Dict[str, Any] = None,
        filter_functions: Iterable[
            Tuple[Union[str, FilterFunction[ValueType, CandidateType, ContextType]], Dict[str, Any]]
        ] = (),
        min_score: float = 0.90,
        overrides: Union[
            InheritedKeysDict[ContextType, ValueType, CandidateType], Dict[ValueType, CandidateType]
        ] = None,
        unmapped_values_action: ActionLevel.ParseType = ActionLevel.IGNORE,
        unknown_user_override_action: ActionLevel.ParseType = ActionLevel.RAISE,
        cardinality: Optional[Cardinality.ParseType] = Cardinality.ManyToOne,
        verbose_logging: bool = False,
    ) -> None:
        self._score = get_by_full_name(score_function, sf) if isinstance(score_function, str) else score_function
        self._score_kwargs = score_function_kwargs or {}
        self._min_score = min_score
        self._overrides: Union[
            InheritedKeysDict[ContextType, ValueType, CandidateType], Dict[ValueType, CandidateType]
        ] = (overrides if isinstance(overrides, InheritedKeysDict) else (overrides or {}))
        self._unmapped_action: ActionLevel = ActionLevel.verify(unmapped_values_action)
        self._bad_candidate_action: ActionLevel = ActionLevel.verify(unknown_user_override_action)
        self._cardinality = None if cardinality is None else Cardinality.parse(cardinality, strict=True)
        self._filters: List[Tuple[FilterFunction[ValueType, CandidateType, ContextType], Dict[str, Any]]] = [
            ((get_by_full_name(func, mf) if isinstance(func, str) else func), kwargs)
            for func, kwargs in filter_functions
        ]
        self._verbose = verbose_logging
        self._logger = logging.getLogger(__package__).getChild("Mapper")  # This will almost always be overwritten

    def apply(
        self,
        values: Iterable[ValueType],
        candidates: Iterable[CandidateType],
        context: ContextType = None,
        override_function: UserOverrideFunction[ValueType, CandidateType, ContextType] = None,
        **kwargs: Any,
    ) -> DirectionalMapping[ValueType, CandidateType]:
        """Map values to candidates.

        Args:
            values: Iterable of elements to match to candidates.
            candidates: Iterable of candidates to match with `value`. Duplicate elements will be discarded.
            context: Context in which mapping is being done. Required when using context-sensitive overrides.
            override_function: A callable that takes inputs ``(value, candidates, context)`` that returns either
                ``None`` (let the regular mapping logic decide) or one of the `candidates`. How non-candidates returned
                is handled is determined by the :attr:`unknown_user_override_action` property.
            **kwargs: Runtime keyword arguments for score and filter functions. May be used to add information which is
                not known when the ``Mapper`` is initialized.

        Returns:
            A :class:`.DirectionalMapping` on the form ``{value: [matched_candidates, ...]}``. May be turned into a
            plain dict ``{value: candidate}`` by using the :meth:`.DirectionalMapping.flatten` function (only if
            :attr:`.DirectionalMapping.cardinality` is of type :attr:`.Cardinality.one_right`).

        Raises:
            MappingError: If any values failed to match and ``unmapped_values_action='raise'``.
            BadFilterError: If a filter returns candidates that are not a subset of the original candidates.
            UserMappingError: If `override_function` returns an unknown candidate and
                ``unknown_user_override_action != 'ignore'``
            ValueError: If passing ``context=None`` (the default) when  :attr:`context_sensitive_overrides` is ``True``.
        """
        start = perf_counter()

        if isinstance(self._overrides, InheritedKeysDict) and context is None:
            raise ValueError("Must pass a context in context-sensitive mode.")

        if self.verbose_logging:
            with enable_verbose_debug_messages():
                scores = self.compute_scores(values, candidates, context, override_function, **kwargs)
        else:
            scores = self.compute_scores(values, candidates, context, override_function, **kwargs)  # pragma: no cover

        dm: DirectionalMapping[ValueType, CandidateType] = self.to_directional_mapping(scores)

        unmapped = set(scores.index[~np.isinf(scores).all(axis=1)]).difference(dm.left)
        if unmapped:
            extra = f" in {context=}" if context else ""
            candidates = set(scores)
            self._report_unmapped(f"Could not map {unmapped}{extra} to any of {candidates=}.")

        verbose_logger = self._get_verbose_logger()
        if verbose_logger.isEnabledFor(logging.DEBUG):
            values = list(scores.index)
            candidates = list(scores.columns)
            cardinality = "automatic" if self.cardinality is None else self.cardinality.name

            l2r = dm.left_to_right
            matches = " Matches:\n" + "\n".join(
                (f"    {v!r} -> {repr(l2r[v]) if v in l2r else '<no matches>'}" for v in values)
            )

            verbose_logger.debug(
                f"Mapping with {cardinality=} completed for {values}x{candidates} in {format_perf_counter(start)}."
                f"{matches}\nMatched {len(dm.left)}/{len(values)} values with {len(dm.right)} different candidates."
            )

        return dm

    def _report_unmapped(self, msg: str) -> None:
        if self.unmapped_values_action is ActionLevel.RAISE:
            msg += (
                "\nHint: set "
                f"unmapped_values_action='{ActionLevel.WARN.value}' or "
                f"unmapped_values_action='{ActionLevel.IGNORE.value}' "
                "to allow unmapped values."
            )
            self.logger.error(msg)
            raise MappingError(msg)
        elif self.unmapped_values_action is ActionLevel.WARN:
            self.logger.warning(msg)
            warnings.warn(msg, MappingWarning, stacklevel=2)
        else:
            self._get_verbose_logger().debug(msg)

    def compute_scores(
        self,
        values: Iterable[ValueType],
        candidates: Iterable[CandidateType],
        context: ContextType = None,
        override_function: UserOverrideFunction[ValueType, CandidateType, ContextType] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Compute likeness scores.

        Args:
            values: Iterable of elements to match to candidates.
            candidates: Iterable of candidates to match with `value`. Duplicate elements will be discarded.
            context: Context in which mapping is being done.
            override_function: A callable that takes inputs ``(value, candidates, context)`` that returns either
                ``None`` (let the regular mapping logic decide) or one of the `candidates`. How non-candidates returned
                is handled is determined by the :attr:`unknown_user_override_action` property.
            **kwargs: Runtime keyword arguments for score and filter functions. May be used to add information which is
                not known when the ``Mapper`` is initialized.

        Returns:
            A ``DataFrame`` of value-candidate match scores, with ``DataFrame.index=values`` and
            ``DataFrame.columns=candidates``.

        Raises:
            BadFilterError: If a filter returns candidates that are not a subset of the original candidates.
            UserMappingError: If `override_function` returns an unknown candidate and
                ``unknown_user_override_action != 'ignore'``
        """
        start = perf_counter()

        scores = pd.DataFrame(
            data=-np.inf,
            columns=pd.Index(list(candidates), name="candidates").drop_duplicates(),
            index=pd.Index(list(values), name="values").drop_duplicates(),
            dtype=float,
        )

        extra = f" in {context=}" if context else ""

        if scores.empty:
            if self.logger.isEnabledFor(logging.DEBUG):
                candidates = list(scores.columns)
                values = list(scores.index)
                end = "" if (values or candidates) else ", but got neither"
                self.logger.warning(
                    f"Abort mapping{extra} of {values}x{candidates}. Both values and candidates must be given{end}."
                )
            return scores

        if self.logger.isEnabledFor(logging.DEBUG):
            candidates = list(scores.columns)
            values = list(scores.index)
            self.logger.debug(f"Begin computing match scores{extra} for {values}x{candidates} using {self._score}.")

        unmapped_values = self._handle_overrides(scores, context, override_function)
        if not unmapped_values:
            return scores

        verbose_logger = self._get_verbose_logger()
        for value in unmapped_values:
            filtered_candidates = self._apply_filters(value, scores.columns, context, kwargs)
            if not filtered_candidates:
                continue

            if verbose_logger.isEnabledFor(logging.DEBUG):
                verbose_logger.debug(f"Compute match scores for {value=}.")
            scores_for_value = self._score(value, filtered_candidates, context, **self._score_kwargs, **kwargs)
            for score, candidate in zip(scores_for_value, filtered_candidates):
                scores.loc[value, candidate] = score

        if verbose_logger.isEnabledFor(logging.DEBUG):
            verbose_logger.debug(
                f"Computed {len(scores.index)}x{len(scores.columns)} "
                f"match scores in {format_perf_counter(start)}:\n{scores.to_string()}"
            )
        return scores

    def to_directional_mapping(
        self,
        scores: pd.DataFrame,
    ) -> DirectionalMapping[ValueType, CandidateType]:
        """Create a ``DirectionalMapping`` from match scores.

        Args:
            scores: A score matrix, where ``scores.index`` are values and ``score.columns`` are treated as the
                candidates.

        Returns:
            A ``DirectionalMapping``.

        See Also:
            :meth:`.MatchScores.to_directional_mapping`
        """
        return MatchScores(scores, self._min_score, self._get_verbose_logger()).to_directional_mapping(self.cardinality)

    def _get_verbose_logger(self) -> logging.Logger:
        logger = self.logger.getChild("verbose")
        logger.disabled = not (FORCE_VERBOSE or self.verbose_logging)
        return logger

    @property
    def cardinality(self) -> Optional[Cardinality]:
        """Return upper cardinality bound during mapping."""
        return self._cardinality

    @property
    def unmapped_values_action(self) -> ActionLevel:
        """Return the action to take if mapping fails for any values."""
        return self._unmapped_action

    @property
    def unknown_user_override_action(self) -> ActionLevel:
        """Return the action to take if an override function returns an unknown candidate.

        Unknown candidates, i.e. candidates not in the input `candidates` collection, will not be used unless `'ignore'`
        is chosen. As such, `'ignore'` should rather be interpreted as `'allow'`.

        Returns:
            Action to take if a user-defined override function returns an unknown candidate.
        """
        return self._bad_candidate_action

    @property
    def context_sensitive_overrides(self) -> bool:
        """Return ``True`` if the overrides are context-sensitive."""
        return isinstance(self._overrides, InheritedKeysDict)

    @property
    def verbose_logging(self) -> bool:
        """Return ``True`` if verbose debug-level messages are enabled."""
        return self._verbose

    @property
    def logger(self) -> logging.Logger:
        """Return the ``Logger`` that is used by this instance."""
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger) -> None:
        self._logger = logger

    def _handle_overrides(
        self,
        scores: pd.DataFrame,
        context: Optional[ContextType],
        override_function: Optional[UserOverrideFunction[ValueType, CandidateType, ContextType]],
    ) -> List[ValueType]:
        applied: Dict[ValueType, CandidateType] = {}

        def apply(v: ValueType, oc: CandidateType) -> None:
            scores.loc[v, :] = -np.inf
            scores.loc[v, oc] = np.inf
            unmapped_values.remove(v)
            applied[v] = oc

        unmapped_values = list(scores.index)

        if override_function:
            for value, override_candidate in self._get_function_overrides(
                override_function, scores.index, scores.columns, context
            ):
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug(
                        f"Using override {repr(value)} -> {repr(override_candidate)} returned by {override_function}."
                    )
                apply(value, override_candidate)

        for value, override_candidate in self._get_static_overrides(unmapped_values, context):
            apply(value, override_candidate)

        if self.logger.isEnabledFor(logging.DEBUG) and (self._overrides or override_function is not None):
            num_overrides = len(self._overrides) + int(override_function is not None)
            result = f"and found {len(applied)} matches={applied} in" if applied else "but none were a match for"
            done = "All values mapped by overrides. " if (not unmapped_values and applied) else ""
            self.logger.debug(
                f"{done}Applied {num_overrides} overrides, {result} the given values={list(scores.index)}."
            )

        return unmapped_values

    def _get_static_overrides(
        self,
        values: Iterable[ValueType],
        context: Optional[ContextType],
    ) -> List[Tuple[ValueType, CandidateType]]:
        if not self._overrides:
            return []

        # The assertions are redundant (checked earlier), but makes mypy happy without type-ignores.
        if self.context_sensitive_overrides:
            assert isinstance(self._overrides, InheritedKeysDict), "Makes mypy happy."  # noqa: S101
            assert context is not None, "Makes mypy happy."  # noqa: S101
            overrides = self._overrides.get(context, {})
        else:
            assert not isinstance(self._overrides, InheritedKeysDict), "Makes mypy happy."  # noqa: S101
            assert context is None, "Makes mypy happy."  # noqa: S101
            overrides = self._overrides

        return [(value, overrides[value]) for value in filter(overrides.__contains__, values)]

    def _get_function_overrides(
        self,
        func: UserOverrideFunction[ValueType, CandidateType, ContextType],
        values: Iterable[ValueType],
        candidates: Iterable[CandidateType],
        context: Optional[ContextType],
    ) -> List[Tuple[ValueType, CandidateType]]:
        candidates = set(candidates)

        ans = []
        for value in values:
            user_override = func(value, candidates, context)
            if user_override is None:
                continue
            if self.unknown_user_override_action is not ActionLevel.IGNORE and user_override not in candidates:
                msg = (
                    f"The user-defined override function {func} returned an unknown candidate={repr(user_override)} for"
                    f" {value=}. If this is intended behaviour, set unknown_user_override_action='ignore' to allow."
                )
                if self.unknown_user_override_action is ActionLevel.RAISE:
                    self.logger.error(msg)
                    raise UserMappingError(msg, value, candidates)
                elif self.unknown_user_override_action is ActionLevel.WARN:
                    msg += " The override has been ignored."
                    self.logger.warning(msg)
                    warnings.warn(msg, UserMappingWarning, stacklevel=2)
                continue

            ans.append((value, user_override))

        return ans

    def _apply_filters(
        self,
        value: ValueType,
        candidates: Iterable[CandidateType],
        context: Optional[ContextType],
        kwargs: Dict[str, Any],
    ) -> Set[CandidateType]:
        candidates = list(candidates)
        filtered_candidates = set(candidates)

        for filter_function, function_kwargs in self._filters:
            filtered_candidates = filter_function(value, filtered_candidates, context, **function_kwargs, **kwargs)

            not_in_original_candidates = filtered_candidates.difference(candidates)
            if not_in_original_candidates:
                raise exceptions.BadFilterError(
                    f"Filter {tname(filter_function)}({value}, candidates, **{kwargs}) created new"
                    f"candidates: {not_in_original_candidates}"
                )

            if not filtered_candidates:
                break

        if self.verbose_logging and self.logger.isEnabledFor(logging.DEBUG) and len(self._filters):
            diff = set(candidates).difference(filtered_candidates)
            removed = f"removing candidates={diff}" if diff else "but did not remove any candidates"
            done = "All candidates removed by filtering. " if not filtered_candidates else ""
            self.logger.debug(f"{done}Applied {len(self._filters)} filters for {value=}, {removed}.")

        return filtered_candidates

    def __repr__(self) -> str:
        score = self._score
        return f"{tname(self)}({score=} >= {self._min_score}, {len(self._filters)} filters)"

    def copy(self, **overrides: Any) -> "Mapper[ValueType, CandidateType, ContextType]":
        """Make a copy of this ``Mapper``.

        Args:
            overrides: Keyword arguments to use when instantiating the copy. Options that aren't given will be taken
                from the current instance. See the :class:`Mapper` class documentation for possible choices.

        Returns:
            A copy of this ``Mapper`` with `overrides` applied.
        """
        kwargs: Dict[str, Any] = {
            "score_function": self._score,
            "min_score": self._min_score,
            "unmapped_values_action": self.unmapped_values_action,
            "unknown_user_override_action": self.unknown_user_override_action,
            "cardinality": self.cardinality,
            "verbose_logging": self.verbose_logging,
            **overrides,
        }

        if "score_function_kwargs" not in kwargs:
            kwargs["score_function_kwargs"] = self._score_kwargs.copy()

        if "filter_functions" not in kwargs:
            kwargs["filter_functions"] = [(func, func_kwargs.copy()) for func, func_kwargs in self._filters]

        if "overrides" not in kwargs:
            kwargs["overrides"] = self._overrides.copy()

        return Mapper(**kwargs)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Mapper):
            return False

        return all(
            (
                self._score == other._score,
                self._score_kwargs == other._score_kwargs,
                self._filters == other._filters,
                self._min_score == other._min_score,
                self._overrides == other._overrides,
                self._unmapped_action == other._unmapped_action,
                self._bad_candidate_action == other._bad_candidate_action,
                self._cardinality == other._cardinality,
                self._verbose == other._verbose,
            )
        )



import json
import sys
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict


def _initialize_versions() -> Dict[str, str]:
    from pandas import __version__ as pandas_version
    from rics import __version__ as rics
    from sqlalchemy import __version__ as sqlalchemy

    from id_translation import __version__ as id_translation

    ans = dict(
        rics=rics,
        id_translation=id_translation,
        sqlalchemy=sqlalchemy,
        pandas=pandas_version,
    )

    if sys.version_info < (3, 11):  # pragma: no cover
        import tomli

        ans["tomli"] = tomli.__version__
    return ans


class BaseMetadata(ABC):
    """Base implementation for Metadata types.

    Args:
        versions: Top-level dependency versions.
        created: The time at which the metadata was originally created.
    """

    def __init__(self, versions: Dict[str, str] = None, created: datetime = None) -> None:
        self.versions = versions or _initialize_versions()
        self.created = created or datetime.now()

    @abstractmethod
    def _serialize(self, to_json: Dict[str, Any]) -> Dict[str, Any]:
        """Turn `to_json` into JSON-serializable types."""

    @classmethod
    @abstractmethod
    def _deserialize(cls, from_json: Dict[str, Any]) -> Dict[str, Any]:
        """Turn `from_json` into desired types."""

    @abstractmethod
    def _is_equivalent(self, other: "BaseMetadata") -> str:
        """Implementation-specific equivalence check."""

    @abstractmethod
    def _log_reject(self, msg: str) -> None:
        """Log metadata reject event. Subclasses should format the `slug` key."""

    @abstractmethod
    def _log_accept(self, msg: str) -> None:
        """Log metadata accept event. Subclasses should format the `kind` key."""

    def is_equivalent(self, other: "BaseMetadata") -> str:
        """Equivalency status."""
        if not isinstance(other, self.__class__):
            return f"Expected class={self.__class__.__name__} but got {other.__class__}"

        for package, version in self.versions.items():
            other_version = other.versions.get(package)
            if other_version != version:
                return f"Expected {package}=={version!r} (your environment) but got {package}=={other_version!r}"

        return self._is_equivalent(other)

    def to_json(self) -> str:
        """Get a JSON representation of this ``BaseMetadata``."""
        raw = self.__dict__.copy()
        kwargs = dict(
            versions=raw.pop("versions"),
            created=raw.pop("created").isoformat(),
        )
        kwargs.update(self._serialize(raw))
        assert not raw, f"Not serialized: {raw}."  # noqa:  S101
        return json.dumps(kwargs, indent=True)

    @classmethod
    def from_json(cls, s: str) -> "BaseMetadata":
        """Create ``BaseMetadata`` from a JSON string `s`."""
        raw = json.loads(s)

        kwargs = dict(
            versions=raw.pop("versions"),
            created=datetime.fromisoformat(raw.pop("created")),
        )
        kwargs.update(cls._deserialize(raw))

        assert not raw, f"Not deserialized: {raw}."  # noqa:  S101
        return cls(**kwargs)

    def use_cached(self, metadata_path: Path, max_age: timedelta) -> bool:
        """Check status of stored metadata config based a desired configuration ``self``.

        Args:
            metadata_path: Path of stored metadata.
            max_age: Maximum age of stored metadata.

        Returns:
            ``True`` if the cached instance is still viable.
        """

        def log_reject(reason: str) -> None:
            self._log_reject(f"{{slug}}; {reason}. Metadata path='{metadata_path}'.")

        if not metadata_path.exists():
            log_reject("no cache metadata found")
            return False

        stored_config = self.from_json(metadata_path.read_text())

        reason_not_equivalent = self.is_equivalent(stored_config)
        if reason_not_equivalent:
            log_reject(f"cached instance is not equivalent. {reason_not_equivalent}")
            return False

        expires_at = stored_config.created + max_age
        offset = timedelta(abs(datetime.now() - expires_at).total_seconds() // 60)

        if expires_at <= stored_config.created:
            log_reject(f"cache expired at {expires_at} ({offset} ago)")
            return False

        self._log_accept(f"Accept cached {{kind}} in '{metadata_path.parent}'. Expires at {expires_at} (in {offset}).")
        return True



from typing import Iterable, List

from rics.misc import tname

from . import parse_format_string
from .types import FormatType, PlaceholdersTuple


class Format:
    r"""Format specification for translations strings.

    Translation formats are similar to regular f-strings, with two important exceptions:

        1. Positional placeholders (``'{}'``) may not be used; correct form is ``'{placeholder-name}'``.
        2. Substrings surrounded by ``'[]'`` denote an optional element. Optional elements..

           * `Must` contain at least one placeholder.
           * Are rendered only if `all` of its placeholders are defined.
           * Are rendered `without` delimiting brackets.

     .. hint::

        Literal angle brackets are added by doubling the wanted character, as for ``'{'`` and ``'}'`` in plain Python
        f-strings. For example, ``'[['`` will render a ``'['``-literal.

    Args:
        fmt: A translation fstring.

    Examples:
        A format string with an optionl element.

        >>> from id_translation.offline import Format
        >>> fmt = Format('{id}:{name}[, nice={is_nice}]')

        The ``Format`` class when used directly only returns required placeholders by default..

        >>> fmt.fstring(), fmt.fstring().format(id=0, name='Tarzan')
        ('{id}:{name}', '0:Tarzan')

        ..but the `placeholders` attribute can be used to retrieve all placeholders, required and optional:

        >>> fmt.placeholders
        ('id', 'name', 'is_nice')
        >>> fmt.fstring(fmt.placeholders), fmt.fstring(fmt.placeholders).format(id=1, name='Morris', is_nice=True)
        ('{id}:{name}, nice={is_nice}', '1:Morris, nice=True')
    """

    def __init__(self, fmt: str) -> None:
        self._fmt = fmt
        self._elements: List[parse_format_string.Element] = self._parse_format_string(fmt)

    def fstring(self, placeholders: Iterable[str] = None, positional: bool = False) -> str:
        """Create a format string for the given placeholders.

        Args:
            placeholders: Keys to keep. Passing ``None`` is equivalent to passing :attr:`required_placeholders`.
            positional: If ``True``, remove names to return a positional fstring.

        Returns:
            An fstring with optional elements removed unless included in `placeholders`.

        Raises:
            KeyError: If required placeholders are missing.
        """
        placeholders = placeholders or self.required_placeholders
        missing_required_placeholders = set(self.required_placeholders).difference(placeholders)
        if missing_required_placeholders:
            raise KeyError(f"Required key(s) {missing_required_placeholders} missing from {placeholders=}.")

        return self._make_fstring(placeholders, positional=positional)

    def _make_fstring(self, placeholders: Iterable[str], positional: bool) -> str:
        def predicate(e: parse_format_string.Element) -> bool:
            return e.required or set(placeholders).issuperset(e.placeholders)

        return "".join(e.positional_part if positional else e.part for e in filter(predicate, self._elements))

    @staticmethod
    def parse(fmt: FormatType) -> "Format":
        """Parse a format.

        Args:
            fmt: Input to parse.

        Returns:
            A ``Format`` instance.
        """
        return fmt if isinstance(fmt, Format) else Format(fmt)

    @property
    def placeholders(self) -> PlaceholdersTuple:
        """All placeholders in the order in which they appear."""
        return self._extract_placeholders(self._elements)

    @property
    def required_placeholders(self) -> PlaceholdersTuple:
        """All required placeholders in the order in which they appear."""
        return self._extract_placeholders(filter(lambda e: e.required, self._elements))

    @property
    def optional_placeholders(self) -> PlaceholdersTuple:  # pragma: no cover
        """All optional placeholders in the order in which they appear."""
        return self._extract_placeholders(filter(lambda e: not e.required, self._elements))

    @staticmethod
    def _extract_placeholders(elements: Iterable[parse_format_string.Element]) -> PlaceholdersTuple:
        ans = []
        for e in elements:
            ans.extend(e.placeholders)
        return tuple(ans)

    @classmethod
    def _parse_format_string(cls, format_string: str) -> List[parse_format_string.Element]:
        return parse_format_string.get_elements(format_string)

    def __str__(self) -> str:
        def repr_part(e: parse_format_string.Element) -> str:
            s = e.part.replace("[", "[[").replace("]", "]]")
            return s if e.required else f"[{s}]"

        return "".join(map(repr_part, self._elements))

    def __repr__(self) -> str:
        return f"{tname(self)}({self.__str__()!r})"



"""Types used for mapping."""
from typing import Callable, Dict, Hashable, Iterable, Optional, Set, Tuple, TypeVar, Union

HL = TypeVar("HL", bound=Hashable)
"""Hashable type on the left side of a directional relationship."""
HR = TypeVar("HR", bound=Hashable)
"""Hashable type on the right side of a directional relationship."""
LeftToRight = Dict[HL, Tuple[HR, ...]]
"""A left-to-right mapping."""
RightToLeft = Dict[HR, Tuple[HL, ...]]
"""A right-to-left mapping."""

ValueType = TypeVar("ValueType", bound=Hashable)
"""A type of item being mapped."""
CandidateType = TypeVar("CandidateType", bound=Hashable)
"""A type of item being mapped."""
MatchTuple = Tuple[CandidateType, ...]
"""A tuple of candidates matched to a value."""
ContextType = TypeVar("ContextType", bound=Hashable)
"""Type of context in which mapping is being performed."""

UserOverrideFunction = Callable[[ValueType, Set[CandidateType], Optional[ContextType]], Optional[CandidateType]]
"""Signature for a user-defined override function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which scoring is being performed.

Returns:
    Either ``None`` (let regular logic decide) or a single candidate `c` in `candidates`.
"""

ScoreFunction = Callable[[ValueType, Iterable[CandidateType], Optional[ContextType]], Iterable[float]]
"""Signature for a likeness score function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which scoring is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Yields:
    A score for each candidate `c` in `candidates`.
"""

AliasFunction = Callable[
    [ValueType, Iterable[CandidateType], Optional[ContextType]], Tuple[ValueType, Iterable[CandidateType]]
]
"""Signature for an alias function for heuristic scoring.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which mapping is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Returns:
    A tuple (name, candidates) with applied heuristics to increase (or decrease) score as desired.
"""

FilterFunction = Callable[[ValueType, Iterable[CandidateType], Optional[ContextType]], Set[CandidateType]]
"""Signature for a filter function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which filtering is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Returns:
    A subset of candidates to keep.
"""

HeuristicsTypes = Union[
    AliasFunction[ValueType, CandidateType, ContextType], FilterFunction[ValueType, CandidateType, ContextType]
]
"""Types that may be interpreted as a score function heuristic."""



"""Test implementations."""
from typing import Any, Collection, Dict, Iterable, List, Tuple

import pandas as pd

from .fetching import Fetcher as _Fetcher
from .fetching.types import IdsToFetch as _IdsToFetch
from .mapping import DirectionalMapping as _DirectionalMapping, Mapper as _Mapper
from .mapping.types import ContextType, UserOverrideFunction, ValueType
from .offline.types import (
    PlaceholderTranslations as _PlaceholderTranslations,
    SourcePlaceholderTranslations as _SourcePlaceholderTranslations,
)
from .types import IdType, SourceType


class TestMapper(_Mapper[ValueType, ValueType, ContextType]):
    """Dummy ``Mapper`` implementation."""

    def apply(
        self,
        values: Iterable[ValueType],
        candidates: Iterable[ValueType],
        context: ContextType = None,
        override_function: UserOverrideFunction[ValueType, ValueType, ContextType] = None,
        **kwargs: Any,
    ) -> _DirectionalMapping[ValueType, ValueType]:
        """Map values to themselves, unless `override_function` is given."""
        values = set(values)

        left_to_right: Dict[ValueType, Tuple[ValueType, ...]] = {v: (v,) for v in values}

        if override_function:
            candidates = set(candidates)
            for v in values:
                user_override = override_function(v, candidates, context)
                if user_override is not None:
                    left_to_right[v] = (user_override,)

        return _DirectionalMapping(left_to_right=left_to_right)


class TestFetcher(_Fetcher[SourceType, IdType]):
    """Dummy ``Fetcher`` implementation.

    A "happy path" fetcher implementation for testing purposes. Returns generated names for all IDs and placeholders,
    so translation retrieval will never fail when using this fetcher.
    """

    def __init__(self, sources: Collection[SourceType] = None) -> None:
        self._sources = set(sources or [])

    @property
    def allow_fetch_all(self) -> bool:
        return False  # pragma: no cover

    @property
    def online(self) -> bool:
        return False  # pragma: no cover

    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        return {source: [] for source in self._sources}

    def fetch(
        self,
        ids_to_fetch: Iterable[_IdsToFetch[SourceType, IdType]],
        placeholders: Iterable[str] = (),
        required: Iterable[str] = (),
    ) -> _SourcePlaceholderTranslations[SourceType]:
        """Return generated translations for all IDs and placeholders."""
        return {itf.source: self._generate_data(itf, list(placeholders)) for itf in ids_to_fetch}

    @staticmethod
    def _generate_data(
        itf: _IdsToFetch[SourceType, IdType], placeholders: List[str]
    ) -> _PlaceholderTranslations[SourceType]:
        if itf.ids is None:
            raise NotImplementedError

        ids = list(itf.ids)
        df = pd.DataFrame([[f"{p}-of-{idx}" for p in placeholders] for idx in ids], columns=placeholders)
        df["id"] = ids
        return _PlaceholderTranslations.make(itf.source, df)

    def fetch_all(
        self, placeholders: Iterable[str] = (), required: Iterable[str] = ()
    ) -> _SourcePlaceholderTranslations[SourceType]:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"TestFetcher(sources={repr(self._sources or None)})"



"""Functions that remove candidates."""
from __future__ import annotations

import logging
import re
from typing import Collection, Iterable, List, Literal, Optional, Set, Tuple, Union

WhereOptions = Literal["name", "context", "candidate"]
WHERE_OPTIONS = ("name", "candidate", "context")
WhereArg = Union[WhereOptions, Iterable[WhereOptions]]
"""Determines how where matches must be found during filtering operations."""

VERBOSE: bool = False
"""If ``True`` enable optional DEBUG-level log messages on each heuristic function invocation.

Notes:
    Not all functions have verbose messages.
"""
LOGGER = logging.getLogger(__package__).getChild("verbose").getChild("filter_functions")


def require_regex_match(
    name: str,
    candidates: Iterable[str],
    context: Optional[str],
    regex: Union[str, re.Pattern[str]],
    where: WhereArg,
    keep_if_match: bool = True,
    purpose: str = "matching",
) -> Set[str]:
    """Require a regex match in `name`, `context`, and/or `candidates`.

    Args:
        name: A name.
        candidates: Potential matches for `name`.
        context: Context in which the function is being called.
        regex: A regex pattern to pass to :py:func:`re.compile`.
        where: Which of ('name', 'candidate', 'context') to match in.
        keep_if_match: If ``False``, require that `regex` does _not_ match to keep candidates.
        purpose: A purpose-string used for logging.

    Returns:
        Approved candidates.

    Raises:
        ValueError: If `where` contains `'context'` when `context` is not given.

    See Also:
        The :meth:`banned_substring` method.
    """
    where = _parse_where_args(where)

    pattern = re.compile(regex, flags=re.IGNORECASE) if isinstance(regex, str) else regex
    logger = LOGGER.getChild("require_regex_match")
    candidates = set(candidates)

    # Short-circuit full refusal
    if "name" in where:
        match = pattern.match(name)
        if keep_if_match and not match:
            if VERBOSE:
                logger.debug(f"Refuse {purpose} for {name=}: Does not match {pattern=}.")
            return set()
        if match and not keep_if_match:
            if VERBOSE:
                logger.debug(f"Refuse {purpose} for {name=}: Matches {pattern=}.")
            return set()

    if "context" in where:
        if context is None:
            raise ValueError(f"No context given but 'context' was found in {where=}.")  # pragma: no cover

        match = pattern.match(context)
        if keep_if_match and not match:
            if VERBOSE:
                logger.debug(f"Refuse {purpose} for {context=}: Does not match {pattern=}.")
            return set()
        if match and not keep_if_match:
            if VERBOSE:
                logger.debug(f"Refuse {purpose} for {context=}: Matches {pattern=}.")
            return set()

    if "candidate" not in where:
        return candidates

    kept: List[str] = []
    rejected: List[str] = []
    for cand in candidates:
        lst = kept if (bool(pattern.match(cand)) is keep_if_match) else rejected
        lst.append(cand)

    if VERBOSE and logger.isEnabledFor(logging.DEBUG):
        logger.debug(f"Filtering with {keep_if_match=} and {pattern=}; kept {sorted(kept)}, rejected {rejected}.")

    return set(kept)


def banned_substring(
    name: str,
    candidates: Iterable[str],
    context: Optional[str],
    substrings: Collection[str],
    where: WhereArg,
) -> Set[str]:
    """Prevent mapping if banned substrings are found.

    Matching on `name` or `context` halts all mapping. Matching candidates excludes only those candidates.

    Args:
        name: An element to find matches for.
        candidates: Potential matches for `name` (not used).
        context: Context in which the function is being called.
        substrings: Substrings which may not be present in `name`.
        where: Which of ('name', 'candidate', 'context') to match in. Empty=all.

    Returns:
        Approved candidates.

    See Also:
        The :meth:`require_regex_match` method, which performs the actual work.
    """
    where = _parse_where_args(where)

    remaining = set(candidates)

    for subs in substrings:
        if not remaining:
            break

        matches = require_regex_match(
            name,
            remaining,
            context,
            regex=re.compile(f".*{subs}.*", flags=re.IGNORECASE),
            where=where,
            keep_if_match=False,
        )
        remaining = remaining.intersection(matches)

    return remaining


def _parse_where_args(args: WhereArg) -> Tuple[WhereOptions, ...]:
    if not args:
        raise ValueError(f"At least one of {WHERE_OPTIONS} must be given.")

    args_tuple = tuple((args,) if isinstance(args, str) else tuple(args))
    for where in args_tuple:
        if where not in WHERE_OPTIONS:
            raise ValueError(f"Bad where-argument {repr(args)}; {where=} not in {WHERE_OPTIONS}.")
    return args_tuple



"""Data structure I  exceptions."""

from typing import Any, Type


class DataStructureIOError(RuntimeError):
    """Base class for IO exceptions."""


class UntranslatableTypeError(DataStructureIOError):
    """Exception indicating that a type cannot be translated.

    Args:
        t: A type.
    """

    def __init__(self, t: Type[Any]) -> None:
        super().__init__(f"Type {t} cannot be translated.")


class NotInplaceTranslatableError(DataStructureIOError):
    """Exception indicating that a type cannot be translated in-place.

    Args:
        arg: Something that can't be translated inplace.
    """

    def __init__(self, arg: Any) -> None:
        super().__init__(f"Inplace translation not possible or implemented for type: {type(arg)}")



"""Mapping errors."""
from typing import Any, Set


class MappingError(ValueError):
    """Something failed to map."""

    def __init__(self, msg: str) -> None:
        super().__init__(
            msg + "\n\nFor help, please refer to the "
            "https://rics.readthedocs.io/en/stable/documentation/mapping-primer.html page."
        )


class ScoringDisabledError(MappingError):
    """Indicates that the scoring logic has been disabled. Raised by :func:`.score_functions.disabled`."""

    def __init__(self, value: Any, candidates: Any, context: Any) -> None:
        super().__init__(
            "Scoring disabled; the Mapper is working in override-only mode. Please add an override "
            f"for {value=} in {context=} in order to map it to an appropriate candidate."
        )
        self.value = value
        self.candidates = candidates
        self.context = context


class AmbiguousScoreError(MappingError):
    """Indicates that the scoring logic has produces ambiguous scores."""

    def __init__(self, kind: str, key: Any, match0: Any, match1: Any, cardinality: str, scores: str) -> None:
        hint = f"\n{scores}\nInspect the matrix above for details. You may wish to use a different scoring method."
        super().__init__(
            f"Ambiguous mapping of {kind}={key!r}; matches ({match0}) and ({match1}) "
            f"are in conflict since {cardinality=}.{hint}"
        )


class UserMappingError(MappingError):
    """A user-defined mapping function did something forbidden."""

    def __init__(self, msg: str, value: Any, candidates: Set[Any]) -> None:
        super().__init__(msg)
        self.value = value
        self.candidates = candidates


class CardinalityError(MappingError):
    """Base class for cardinality issues."""


class MappingWarning(UserWarning):
    """Something failed to map."""


class UserMappingWarning(MappingWarning):
    """A user-defined mapping function did something strange."""


class BadFilterError(MappingError):
    """Invalid filter."""



"""Mapping implementations for matching groups of elements.

For and introduction to mapping, see :ref:`mapping-primer`.
"""
import logging as _logging

from ._cardinality import Cardinality
from ._directional_mapping import DirectionalMapping
from ._heuristic_score import HeuristicScore
from ._mapper import Mapper

__all__ = [
    "Cardinality",
    "HeuristicScore",
    "DirectionalMapping",
    "Mapper",
]

VERBOSE_LOGGER = _logging.getLogger(__package__).getChild("verbose")
"""Verbose logger. Only logs messages on the ``DEBUG`` level. Disabled by default."""
VERBOSE_LOGGER.disabled = True
